# Copyright (C) 2414 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import struct
import typing as t

from ..commons import swap_words
from .commons import UNIXFilesystem, UNIXInode, iterate_words
from .unix4fs import V4_INODE_FORMAT, V4_NADDR, UNIX4Filesystem, UNIXInode4


class UNIXInode6(UNIXInode4):

    @classmethod
    def read(cls, fs: "UNIXFilesystem", inode_num: int, buffer: bytes, position: int = 0) -> "UNIXInode":
        self = UNIXInode6(fs)
        self.inode_num = inode_num
        (
            self.flags,  #   1 word  flags
            self.nlinks,  #  1 byte  number of links to file
            self.uid,  #     1 byte  user ID of owner
            self.gid,  #     1 byte  group ID of owner
            sz0,  #          1 byte  high byte of 24-bit size
            sz1,  #          1 word  low word of 24-bit size
            addr,  #         8 words block numbers or device numbers
            self.atime,  #   1 long  time of last access
            self.mtime,  #   1 long  time of last modification
        ) = struct.unpack_from(V4_INODE_FORMAT, buffer, position)
        self.addr = struct.unpack_from(f"{V4_NADDR}H", addr)  # type: ignore
        self.size = (sz0 << 16) + sz1
        self.atime = swap_words(self.atime)
        self.mtime = swap_words(self.mtime)
        return self

    @property
    def is_huge(self) -> bool:
        """
        Extra-large files are not marked by any flag, but only by having addr[7] non-zero
        """
        return self.is_large and (self.addr[V4_NADDR - 1] != 0)

    def blocks(self) -> t.Iterator[int]:
        if self.is_huge:
            # Huge file
            for index, block_number in enumerate(self.addr):
                if block_number == 0:
                    break
                if index < V4_NADDR - 1:
                    indirect_block = self.fs.read_block(block_number)
                    for n in iterate_words(indirect_block):
                        if n == 0:
                            break
                        yield n
                else:
                    double_indirect_block = self.fs.read_block(block_number)
                    for d in iterate_words(double_indirect_block):
                        if d == 0:
                            break
                        indirect_block = self.fs.read_block(d)
                        for n in iterate_words(indirect_block):
                            if n == 0:
                                break
                            yield n
        else:
            # Small and large files
            yield from super().blocks()


class UNIX6Filesystem(UNIX4Filesystem):
    """
    UNIX version 6 Filesystem
    """

    fs_name = "unix6"
    fs_description = "UNIX version 6"
    fs_platforms = ["pdp-11", "interdata7"]
    version: int = 6  # UNIX version
    unix_inode_class = UNIXInode6



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import sys
import traceback
import typing as t

from .abstract import AbstractFilesystem
from .apple2.appledosfs import AppleDOSFilesystem
from .apple2.pascalfs import PascalFilesystem
from .apple2.prodosfs import ProDOSFilesystem
from .commons import DECTAPE, DECTAPE_EXT, splitdrive
from .native import NativeFilesystem
from .nova.dgdosdumpfs import DGDOSDumpFilesystem
from .nova.dgdosfs import DGDOSFilesystem
from .nova.dgdosmagtapefs import DGDOSMagTapeFilesystem
from .pdp7.decsysfs import DECSysFilesystem
from .pdp7.unix0fs import UNIX0Filesystem
from .pdp8.cos300fs import COS300Filesystem
from .pdp8.dmsfs import DMSFilesystem
from .pdp8.os8fs import OS8Filesystem
from .pdp8.tss8fs import TSS8Filesystem
from .pdp11.caps11fs import CAPS11Filesystem
from .pdp11.dos11fs import DOS11Filesystem
from .pdp11.dos11magtapefs import DOS11MagTapeFilesystem
from .pdp11.files11fs import Files11Filesystem
from .pdp11.rstsfs import RSTSFilesystem
from .pdp11.rt11fs import RT11Filesystem
from .pdp11.solofs import SOLOFilesystem
from .pdp15.adssfs import ADSSFilesystem
from .pdp15.dos15fs import DOS15Filesystem
from .unix.unix1fs import UNIX1Filesystem
from .unix.unix4fs import UNIX4Filesystem
from .unix.unix5fs import UNIX5Filesystem
from .unix.unix6fs import UNIX6Filesystem
from .unix.unix7fs import UNIX7Filesystem

__all__ = [
    "Volumes",
    "DEFAULT_VOLUME",
    "FILESYSTEMS",
]

DEFAULT_VOLUME = "DK"
SYSTEM_VOLUME = "SY"
FILESYSTEMS: t.Dict[str, t.Type[AbstractFilesystem]] = {
    "caps11": CAPS11Filesystem,
    "caps8": CAPS11Filesystem,
    "dos11": DOS11Filesystem,
    "dos11mt": DOS11MagTapeFilesystem,
    "files11": Files11Filesystem,
    "rt11": RT11Filesystem,
    "solo": SOLOFilesystem,
    "unix0": UNIX0Filesystem,
    "unix1": UNIX1Filesystem,
    "unix4": UNIX4Filesystem,
    "unix5": UNIX5Filesystem,
    "unix6": UNIX6Filesystem,
    "unix7": UNIX7Filesystem,
    "rsts": RSTSFilesystem,
    "os8": OS8Filesystem,
    "dms": DMSFilesystem,
    "tss8": TSS8Filesystem,
    "cos300": COS300Filesystem,
    "cos310": COS300Filesystem,
    "adss": ADSSFilesystem,
    "dos15": DOS15Filesystem,
    "prodos": ProDOSFilesystem,
    "pascal": PascalFilesystem,
    "appledos": AppleDOSFilesystem,
    "decsys": DECSysFilesystem,
    "dgdos": DGDOSFilesystem,
    "dgdosmt": DGDOSMagTapeFilesystem,
    "dgdosdump": DGDOSDumpFilesystem,
}


class Volumes(object):
    """
    Logical Device Names

    SY: System volume, the device from which this program was started
    DK: Default storage volume (initially the same as SY:)
    """

    volumes: t.Dict[str, AbstractFilesystem]  # volume id -> fs
    logical: t.Dict[str, str]  # local id -> volume id
    defdev: str  # Default device, DK

    def __init__(self) -> None:
        self.volumes: t.Dict[str, AbstractFilesystem] = {}
        self.logical: t.Dict[str, str] = {}
        if self._drive_letters():
            # windows
            for letter in self._drive_letters():
                self.volumes[letter] = NativeFilesystem(f"{letter.upper()}:")
            current_drive = os.getcwd().split(":")[0].upper()
            self.defdev = current_drive
            self.logical["SY"] = current_drive
        else:
            # posix
            self.volumes["N"] = NativeFilesystem()
            self.logical[SYSTEM_VOLUME] = "N"
            self.defdev = SYSTEM_VOLUME

    def _drive_letters(self) -> list[str]:
        try:
            import string
            from ctypes import windll  # type: ignore

            drives = []
            bitmask = windll.kernel32.GetLogicalDrives()
            for c in string.ascii_uppercase:
                if bitmask & 1:
                    drives.append(c)
                bitmask >>= 1
            return drives
        except Exception:
            return []

    def canonical_volume(self, volume_id: str, cmd: str = "KMON") -> str:
        """
        Convert a volume id into canonical form
        """
        if not volume_id:
            volume_id = DEFAULT_VOLUME
        else:
            volume_id = volume_id.upper()
            if volume_id.endswith(":"):
                volume_id = volume_id[:-1]
        return volume_id

    def get(self, volume_id: str, cmd: str = "KMON") -> AbstractFilesystem:
        """
        Get a filesystem by volume id
        """
        volume_id = self.canonical_volume(volume_id, cmd=cmd)
        if volume_id == DEFAULT_VOLUME:
            volume_id = self.defdev
        volume_id = self.logical.get(volume_id, volume_id)
        try:
            return self.volumes[volume_id]
        except KeyError:
            raise Exception(f"?{cmd}-F-Illegal volume {volume_id}:")

    def chdir(self, path: str) -> bool:
        """
        Change current directory
        """
        volume_id, fullname = splitdrive(path)
        volume_id = self.canonical_volume(volume_id)
        try:
            fs = self.get(volume_id)
        except Exception:
            return False
        if fullname and not fs.chdir(fullname):
            return False
        if volume_id != DEFAULT_VOLUME:
            self.set_default_volume(volume_id)
        return True

    def get_pwd(self) -> str:
        """
        Get current volume and directory
        """
        try:
            pwd = self.get(self.defdev).get_pwd()
            return f"{self.defdev}:{pwd}"
        except Exception:
            return f"{self.defdev}:???"

    def set_default_volume(self, volume_id: str, cmd: str = "KMON") -> None:
        """
        Set the default volume
        """
        volume_id = self.canonical_volume(volume_id, cmd=cmd)
        if volume_id != DEFAULT_VOLUME:
            self.get(volume_id, cmd=cmd)
            self.defdev = volume_id

    def assign(self, volume_id: str, logical: str, verbose: bool = False, cmd: str = "KMON") -> None:
        """
        Associate a logical device name with a device
        """
        volume_id = self.canonical_volume(volume_id)
        volume_id = self.logical.get(volume_id, volume_id)
        logical = self.canonical_volume(logical)
        if logical == DEFAULT_VOLUME:
            self.set_default_volume(volume_id, cmd=cmd)
        else:
            self.get(volume_id, cmd=cmd)
            self.logical[logical] = volume_id

    def deassign(self, volume_id: str, verbose: bool = False, cmd: str = "KMON") -> None:
        """
        Removes logical device name assignments
        """
        volume_id = self.canonical_volume(volume_id)
        if volume_id == DEFAULT_VOLUME or not volume_id in self.logical:
            raise Exception(f"?{cmd}-W-Logical name not found {volume_id}:")
        del self.logical[volume_id]

    def mount(
        self,
        path: str,
        logical: str,
        fstype: t.Optional[str] = None,
        device_type: t.Optional[str] = None,
        verbose: bool = False,
        cmd: str = "MOUNT",
    ) -> None:
        """
        Mount a file to a logical disk unit
        """
        logical = self.canonical_volume(logical)
        if logical == DEFAULT_VOLUME or not logical:
            raise Exception(f"?{cmd}-F-Illegal volume {logical}:")
        volume_id, fullname = splitdrive(path)
        fs = self.get(volume_id, cmd=cmd)
        kwargs: t.Dict[str, t.Union[bool, str]] = {}
        device_type = device_type or self.guess_device_type(fullname)
        if device_type:
            kwargs["device_type"] = device_type
        try:
            filesystem = FILESYSTEMS.get(fstype or "rt11", RT11Filesystem)
            self.volumes[logical] = filesystem.mount(fs.open_file(fullname), **kwargs)
            sys.stdout.write(f"?{cmd}-I-Disk {path} mounted to {logical}:\n")
        except Exception:
            if verbose:
                traceback.print_exc()
            sys.stdout.write(f"?{cmd}-F-Error mounting {path} to {logical}:\n")

    def dismount(self, volume_id: str, cmd: str = "DISMOUNT") -> None:
        """
        Disassociates a logical disk assignment from a file
        """
        volume_id = self.canonical_volume(volume_id)
        if volume_id == DEFAULT_VOLUME:
            raise Exception(f"?{cmd}-F-Illegal volume {volume_id}:")
        try:
            fs = self.get(volume_id, cmd=cmd)
        except Exception:
            raise Exception(f"?{cmd}-F-Illegal volume {volume_id}:")
        self.volumes = {k: v for k, v in self.volumes.items() if v != fs}

    def guess_device_type(self, path: str) -> t.Optional[str]:
        """
        Guess the filesystem type based on the file extension
        """
        path = path.lower()
        try:
            _, extension = path.split(".", 1)
        except Exception:
            return None
        if extension in DECTAPE_EXT:
            return DECTAPE
        return None



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import math
import os
import struct
import sys
import typing as t
from datetime import date

from ..abstract import AbstractDirectoryEntry, AbstractFile
from ..commons import (
    ASCII,
    IMAGE,
    READ_FILE_FULL,
    TrackSector,
    dump_struct,
    filename_match,
    hex_dump,
)
from ..device.abstract import AbstractDevice
from .abstract import AbstractAppleDiskFilesystem
from .commons import ProDOSFileInfo, decode_apple_single, encode_apple_single
from .disk import SECTOR_SIZE

__all__ = [
    "AppleDOSFile",
    "AppleDOSDirectoryEntry",
    "AppleDOSFilesystem",
]

# Apple II AppleDOS Filesystem
# https://archive.org/details/apple-ii-pascal-1.3/page/n803/mode/2up

FILENAME_LEN = 30  # max filename length

NEXT_TRACK_OFFSET = 0x01
NEXT_SECTOR_OFFSET = 0x02
TRACK_SECTOR_OFFSET = 0x0C

VTOC_TRACK = 17  # Track of VTOC
VTOC_SECTOR = 0  # Sector of VTOC
VTOC_ADDRESS = TrackSector(VTOC_TRACK, VTOC_SECTOR)
VTOC_FORMAT = "<BBBB2sB32sB8sBB2sBBH"
VTOC_BITMAP_OFFSET = struct.calcsize(VTOC_FORMAT)
VTOC_BITMAP_TRACK_SIZE = 4  # 4 bytes per track
VTOC_MAX_TRACKS = (
    SECTOR_SIZE - VTOC_BITMAP_OFFSET
) // VTOC_BITMAP_TRACK_SIZE  # Maximum number of tracks, limited by the bitmap size

DEFAULT_VOLUME_NUMBER = 254  # Default volume number
DEFAULT_DOS_VERSION = 3  # Default DOS version (DOS 3.3)
DEFAULT_DOS_TYPE = 4  # Default DOS type (DOS 3.3)

RESERVED_TRACKS = 3  # Number of reserved tracks
DATA_SECTORS_PER_TRACK_SECTOR_LIST = 122

FILE_DESCRIPTIVE_ENTRY_OFFSET = 0x0B
FILE_DESCRIPTIVE_ENTRY_FORMAT = "<BBB30sH"
FILE_DESCRIPTIVE_ENTRY_SIZE = struct.calcsize(FILE_DESCRIPTIVE_ENTRY_FORMAT)
FILE_DESCRIPTIVE_ENTRY_PER_SECTOR = 7

FILE_TYPE_TEXT = 0x00  # Text file
FILE_TYPE_INTEGER_BASIC = 0x01  # Integer BASIC file
FILE_TYPE_APPLESOFT_BASIC = 0x02  # Applesoft BASIC file
FILE_TYPE_BINARY = 0x04  # Binary file
FILE_TYPE_SPECIAL = 0x08  # Type S file
FILE_TYPE_RELOCABLE = 0x10  # Relocatable object module file
FILE_TYPE_A = 0x20  # New type A file
FILE_TYPE_B = 0x40  # New type B file

LOCKED_FLAG = 0x80
DELETED_TRACK = 0xFF

FILE_TYPES = {
    FILE_TYPE_TEXT: "T",
    FILE_TYPE_INTEGER_BASIC: "I",
    FILE_TYPE_APPLESOFT_BASIC: "A",
    FILE_TYPE_BINARY: "B",
    FILE_TYPE_SPECIAL: "S",
    FILE_TYPE_RELOCABLE: "R",
    FILE_TYPE_A: "a",
    FILE_TYPE_B: "b",
}

PRODOS_TXT_FILE_TYPE = 0x04  # Text file type (mapped to T)
PRODOS_BIN_FILE_TYPE = 0x06  # Binary file type (mapped to B)
PRODOS_INT_FILE_TYPE = 0xFA  # Integer BASIC file type (mapped to I)
PRODOS_BAS_FILE_TYPE = 0xFC  # Applesoft BASIC file type (mapped to A)
PRODOS_REL_FILE_TYPE = 0xFE  # Relocable object file type (mapped to R)

BINARY_FILE_FORMAT = "<HH"  # Address/Length
BASIC_FILE_FORMAT = "<H"  # Length


def appledos_canonical_filename(fullname: t.Optional[str], wildcard: bool = False) -> t.Optional[str]:
    """
    Generate the canonical AppleDOS filename
    """
    if fullname:
        fullname = fullname[:FILENAME_LEN].upper()
    return fullname


def appledos_get_raw_file_type(file_type: t.Optional[str], default: int = FILE_TYPE_TEXT) -> int:
    """
    Get the file type id from a string
    """
    if not file_type:
        return default
    file_type = file_type.upper()
    for file_id, file_str in FILE_TYPES.items():
        if file_str == file_type:
            return file_id
    raise Exception("?KMON-F-Invalid file type specified with option")


def appledos_filename_to_raw_filename(filename: str) -> bytes:
    """
    Convert the filename to a raw filename
    """
    raw_filename = bytes([ord(x) ^ 0x80 for x in filename[:FILENAME_LEN].upper()])
    raw_filename += b"\xa0" * (FILENAME_LEN - len(raw_filename))
    return raw_filename


class AppleDOSFile(AbstractFile):
    entry: "AppleDOSDirectoryEntry"
    file_mode: str
    closed: bool

    def __init__(self, entry: "AppleDOSDirectoryEntry", file_mode: t.Optional[str] = None):
        self.entry = entry
        self.closed = False
        if file_mode is None:
            self.file_mode = ASCII if entry.raw_file_type == FILE_TYPE_TEXT else IMAGE
        elif file_mode is ASCII or file_mode == FILE_TYPES[FILE_TYPE_TEXT]:
            self.file_mode = ASCII
        else:
            self.file_mode = IMAGE

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.get_length()
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        data = bytearray()
        # Get the blocks to be read
        blocks = list(self.entry.blocks())[block_number : block_number + number_of_blocks]
        # Read the blocks
        for disk_block_number in blocks:
            buffer = self.entry.fs.read_sector(disk_block_number)
            data.extend(buffer)
        # Convert to ASCII if needed
        if self.file_mode == ASCII:
            return bytes([0x0A if x == 0x8D else x & 0x7F for x in data])
        else:
            return bytes(data)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        # Convert to ASCII if needed
        if self.file_mode == ASCII:
            buffer = bytes([0x8D if x == 0x0A else x | 0x80 for x in buffer])
        # Get the blocks to be written
        blocks = list(self.entry.blocks())[block_number : block_number + number_of_blocks]
        # Write the blocks
        for i, disk_block_number in enumerate(blocks):
            data = buffer[i * SECTOR_SIZE : (i + 1) * SECTOR_SIZE]
            self.entry.fs.write_sector(data, disk_block_number)

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.entry.get_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return self.entry.get_block_size()

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class AppleDOSVTOC:
    """
    Volume Table Of Contents

    The VTOC contains information about the diskette and
    the bitmap of the free sectors per track.
    It is always located on track 17, sector 0.

    https://archive.org/details/Beneath_Apple_DOS_alt/page/n19/mode/2up

       Field                                Byte of
      Length                                Sector
             +----------------------------+
     1 byte  |        DOS type            | $00
             +----------------------------+
     1 byte  |       catalog track        | $01
             +----------------------------+
     1 byte  |       catalog sector       | $02
             +----------------------------+
     1 byte  |        DOS version         | $03
             +----------------------------+
     2 byte  |        reserved            | $04
             |                            | $05
             +----------------------------+
     1 byte  |       volume number        | $06
             +----------------------------+
     32 byte |        reserved            | $07
             |                            | $26
             +----------------------------+
     1 byte  |  max track/sector pairs    | $27
             +----------------------------+
     8 byte  |        reserved            | $28
             |                            | $2F
             +----------------------------+
     1 byte  |   last track allocated     | $30
             +----------------------------+
     1 byte  |  allocation direction      | $31
             +----------------------------+
     2 byte  |        reserved            | $32
             |                            | $33
             +----------------------------+
     1 byte  |   number of tracks on disk | $34
             +----------------------------+
     1 byte  |   sectors per track        | $35
             +----------------------------+
     2 byte  |   bytes per sector         | $36
             |                            | $37
             +----------------------------+
     4 byte  | bitmap of track 0          | $38
             |                            | $3B
             +----------------------------+
     4 byte  | bitmap of track 1          | $3C
             |                            | $3F
             +----------------------------+
             | ...                        |
             +----------------------------+
             | bitmap of track n          |
             |                            |
             +----------------------------+

    """

    fs: "AppleDOSFilesystem"
    dos_type: int = 0  # DOS type
    catalog_address: TrackSector  # first catalog address
    dos_version: int = 0  # DOS version
    volume_number: int = 0  # volume number
    max_ts_pairs: int = 0  # max track/sector pairs per sector
    last_track_allocated: int = 0  # last track allocated
    allocation_direction: int = 0  # direction of track allocation (+1 or -1)
    number_of_tracks: int = 0  # number of tracks on disk
    sectors_per_track: int = 0  # sectors per track
    bytes_per_sector: int = 0  # bytes per sector
    reserved_1: bytes  # reserver - 2 bytes
    reserved_2: bytes  # reserver - 32 bytes
    reserved_3: bytes  # reserver - 8 bytes
    reserved_4: bytes  # reserver - 2 bytes
    bitmaps: t.List[int]  # bitmap blocks (32 bits per track)

    def __init__(self, fs: "AppleDOSFilesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "AppleDOSFilesystem") -> "AppleDOSVTOC":
        """
        Read the VTOC sector
        """
        self = AppleDOSVTOC(fs)
        buffer = fs.read_sector(VTOC_ADDRESS)
        (
            self.dos_type,  # DOS type
            catalog_track,  # track of catalog sector
            catalog_sector,  # sector of catalog sector
            self.dos_version,  # DOS version
            self.reserved_1,  # reserved
            self.volume_number,  # volume number
            self.reserved_2,  # reserved
            self.max_ts_pairs,  # max track/sector pairs per sector
            self.reserved_3,  # reserved
            self.last_track_allocated,  # last track allocated
            self.allocation_direction,  # allocation direction
            self.reserved_4,  # reserved
            self.number_of_tracks,  # number of tracks on disk
            self.sectors_per_track,  # sectors per track
            self.bytes_per_sector,  # bytes per sector
        ) = struct.unpack_from(VTOC_FORMAT, buffer, 0)
        self.catalog_address = TrackSector(catalog_track, catalog_sector)
        # Read the bitmap blocks
        bitmap_format = f">{self.number_of_tracks}I"
        self.bitmaps = list(struct.unpack_from(bitmap_format, buffer, VTOC_BITMAP_OFFSET))
        return self

    @classmethod
    def create(cls, fs: "AppleDOSFilesystem") -> "AppleDOSVTOC":
        """
        Create the VTOC sector
        """
        self = AppleDOSVTOC(fs)
        # Initialize the VTOC sector
        self.dos_type = DEFAULT_DOS_TYPE  # DOS type
        self.catalog_address = TrackSector(VTOC_TRACK, fs.dev.sectors_per_track - 1)
        self.dos_version = DEFAULT_DOS_VERSION  # DOS version
        self.reserved_1 = b"\0" * 2  # reserved
        self.volume_number = DEFAULT_VOLUME_NUMBER  # volume number
        self.reserved_2 = b"\0" * 32  # reserved
        self.max_ts_pairs = DATA_SECTORS_PER_TRACK_SECTOR_LIST  # max track/sector pairs per sector
        self.reserved_3 = b"\0" * 8  # reserved
        self.last_track_allocated = VTOC_TRACK  # last track allocated
        self.allocation_direction = 255  # allocation direction
        self.reserved_4 = b"\0" * 2  # reserved
        self.number_of_tracks = fs.dev.number_of_tracks  # number of tracks on disk
        if fs.dev.number_of_tracks > VTOC_MAX_TRACKS:
            self.number_of_tracks = VTOC_MAX_TRACKS
            fs.dev.number_of_tracks = self.number_of_tracks
        self.sectors_per_track = fs.dev.sectors_per_track  # sectors per track
        self.bytes_per_sector = SECTOR_SIZE  # bytes per sector
        # Initialize the bitmap
        b = ((1 << self.sectors_per_track) - 1) << (32 - self.sectors_per_track)
        self.bitmaps = [b] * self.number_of_tracks
        # The first 3 tracks are reserved for the bootstrap image
        for track in range(0, RESERVED_TRACKS):
            self.bitmaps[track] = 0
        # The track 17 is reserved for the VTOC and the catalog
        self.bitmaps[VTOC_TRACK] = 0
        return self

    def write(self) -> None:
        """
        Write the VTOC sector
        """
        buffer = bytearray(SECTOR_SIZE)
        struct.pack_into(
            VTOC_FORMAT,
            buffer,
            0,
            self.dos_type,  # DOS type
            self.catalog_address.track,  # track of catalog sector
            self.catalog_address.sector,  # sector of catalog sector
            self.dos_version,  # DOS version
            self.reserved_1,  # reserved
            self.volume_number,  # volume number
            self.reserved_2,  # reserved
            self.max_ts_pairs,  # max track/sector pairs per sector
            self.reserved_3,  # reserved
            self.last_track_allocated,  # last track allocated
            self.allocation_direction,  # allocation direction
            self.reserved_4,  # reserved
            self.number_of_tracks,  # number of tracks on disk
            self.sectors_per_track,  # sectors per track
            self.bytes_per_sector,  # bytes per sector
        )
        # Write the bitmap blocks
        bitmap_format = f">{self.number_of_tracks}I"
        struct.pack_into(bitmap_format, buffer, VTOC_BITMAP_OFFSET, *self.bitmaps)
        self.fs.write_sector(buffer, VTOC_ADDRESS)

    @property
    def is_valid(self) -> bool:
        """
        Check if the VTOC is valid
        """
        return self.bytes_per_sector == SECTOR_SIZE

    @property
    def total_bits(self) -> int:
        """
        Return the bitmap length in bit
        """
        return len(self.bitmaps) * 8

    def is_free(self, address: TrackSector) -> bool:
        """
        Check if the sector is free
        """
        bit_value = self.bitmaps[address.track]
        bit_position = address.sector + (32 - self.sectors_per_track)
        return (bit_value & (1 << bit_position)) != 0

    def set_free(self, address: TrackSector) -> None:
        """
        Mark the sector as free
        """
        bit_position = address.sector + (32 - self.sectors_per_track)
        self.bitmaps[address.track] |= 1 << bit_position

    def set_used(self, address: TrackSector) -> None:
        """
        Mark the sector as used
        """
        bit_position = address.sector + (32 - self.sectors_per_track)
        self.bitmaps[address.track] &= ~(1 << bit_position)

    def allocate(self, size: int) -> t.List[TrackSector]:
        """
        Allocate sectors
        """
        if self.free() < size:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        sectors: t.List[TrackSector] = []
        # Get the last allocated track
        track = self.last_track_allocated
        while size > 0:
            # Calculate the next track
            if self.allocation_direction == 1:
                track += 1
                if track >= self.number_of_tracks:
                    track = 1  # skip track 0
            else:
                track -= 1
                if track <= 0:
                    track = self.number_of_tracks - 1
            # Allocate sectors
            for sector in range(self.sectors_per_track, 0, -1):
                address = TrackSector(track, sector)
                if self.is_free(address):
                    self.set_used(address)
                    sectors.append(address)
                    size -= 1
                if size == 0:
                    break
            if track == self.last_track_allocated:
                break
        # Update the last allocated track
        self.last_track_allocated = track
        if len(sectors) < size:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        return sectors

    def used(self) -> int:
        """
        Count the number of used sectors
        """
        return self.number_of_tracks * self.sectors_per_track - self.free()

    def free(self) -> int:
        """
        Count the number of free sectors
        """
        free = 0
        for block in self.bitmaps:
            free += block.bit_count()
        return free

    def __str__(self) -> str:
        return dump_struct(self.__dict__, exclude=["fs", "bitmaps"])


class AppleDOSCatalog:
    """
    Each catalog sector contains up to 7 file entries.
    Each sector has a track/sector pointer in bytes 01 and 02
    which points to the next catalog sector.
    The last catalog sector has a 0/0 pointer to indicate that there
    are no more catalog sectors in the chain.

       Field                                Byte of
      Length                                Sector
             +----------------------------+
     1 byte  |        reserved            | $00
             |----------------------------|
     1 byte  |       next track           | $01
             |----------------------------|
     1 byte  |       next sector          | $02
             +----------------------------+
             |                            | $03
     7 byte  |        reserved            |
             |                            |
             +----------------------------+
     34 byte |      file entry 1          | $0B
     each    |          ...               |
     238     |      file entry 7          |
     total   +----------------------------+

    https://archive.org/details/Beneath_Apple_DOS_alt/page/n19/mode/2up
    """

    fs: "AppleDOSFilesystem"
    catalog_addresses: t.List[TrackSector]  # catalog sectors addresses
    directory_entries: t.List["AppleDOSDirectoryEntry"]  # 7 file entries per catalog sector

    def __init__(self, fs: "AppleDOSFilesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "AppleDOSFilesystem", catalog_address: t.Optional[TrackSector] = None) -> "AppleDOSCatalog":
        """
        Read the catalog from the disk
        """
        self = cls(fs)
        self.catalog_addresses = []
        self.directory_entries = []
        address = catalog_address if catalog_address is not None else self.fs.catalog_address
        while address.track != 0:
            # Read the catalog sector
            self.catalog_addresses.append(address)
            buffer = self.fs.read_sector(address)
            # Read 7 file entries
            for position in range(FILE_DESCRIPTIVE_ENTRY_OFFSET, SECTOR_SIZE, FILE_DESCRIPTIVE_ENTRY_SIZE):
                entry = AppleDOSDirectoryEntry.read(self.fs, buffer, position)
                self.directory_entries.append(entry)
            address = TrackSector(buffer[NEXT_TRACK_OFFSET], buffer[NEXT_SECTOR_OFFSET])
        return self

    @classmethod
    def create(cls, fs: "AppleDOSFilesystem", vtoc: "AppleDOSVTOC") -> "AppleDOSCatalog":
        """
        Create an empty catalog on the remaining sectors of VTOC track
        """
        self = cls(fs)
        self.catalog_addresses = [
            TrackSector(VTOC_TRACK, sector) for sector in range(self.fs.dev.sectors_per_track - 1, 0, -1)
        ]
        self.directory_entries = [
            AppleDOSDirectoryEntry.create(self.fs)
            for x in range(0, FILE_DESCRIPTIVE_ENTRY_PER_SECTOR * len(self.catalog_addresses))
        ]
        return self

    def write(self) -> None:
        """
        Write the catalog to the disk
        """
        entries = list(self.directory_entries)
        for i in range(0, len(self.catalog_addresses)):
            buffer = bytearray(SECTOR_SIZE)
            # Update the link to next catalog sector
            if i + 1 < len(self.catalog_addresses):
                next_address = self.catalog_addresses[i + 1]
                buffer[NEXT_TRACK_OFFSET] = next_address.track
                buffer[NEXT_SECTOR_OFFSET] = next_address.sector
            else:
                buffer[NEXT_TRACK_OFFSET] = 0
                buffer[NEXT_SECTOR_OFFSET] = 0
            # Update the catalog sector
            for j in range(0, FILE_DESCRIPTIVE_ENTRY_PER_SECTOR):
                if not entries:
                    break
                entry = entries.pop(0)
                entry.write_buffer(buffer, FILE_DESCRIPTIVE_ENTRY_OFFSET + j * FILE_DESCRIPTIVE_ENTRY_SIZE)
            # Write the catalog sector
            self.fs.write_sector(buffer, self.catalog_addresses[i])

    def iterdir(self, include_deleted: bool = False) -> t.Iterator["AppleDOSDirectoryEntry"]:
        """
        Iterate over directory entries
        """
        for entry in self.directory_entries:
            if not entry.is_empty and (include_deleted or not entry.is_deleted):
                yield entry

    def search_empty_entry(self) -> t.Optional["AppleDOSDirectoryEntry"]:
        """
        Search for an empty or deleted catalog entry
        """
        # Search for an empty catalog entry
        for entry in self.directory_entries:
            if entry.is_empty:
                return entry
        # Search for a delete catalog entry
        for entry in self.directory_entries:
            if entry.is_deleted:
                return entry
        return None

    def create_file(
        self,
        vtoc: "AppleDOSVTOC",
        fullname: str,
        number_of_blocks: int,  # length in blocks
        file_type: t.Optional[str],  # optional file type
    ) -> "AppleDOSDirectoryEntry":
        """
        Create a new file
        """
        # Allocate sectors
        trac_sector_list_num = max(math.ceil(number_of_blocks / DATA_SECTORS_PER_TRACK_SECTOR_LIST), 1)
        sectors = vtoc.allocate(number_of_blocks + trac_sector_list_num)
        trac_sector_list_sectors = sectors[:trac_sector_list_num]
        data_sectors = sectors[trac_sector_list_num:]
        # Prepare the track/sector list
        for i in range(trac_sector_list_num):
            buffer = bytearray(SECTOR_SIZE)
            # Update the link to next track/sector list sector
            if i + 1 < trac_sector_list_num:
                next_address = trac_sector_list_sectors[i + 1]
                buffer[NEXT_TRACK_OFFSET] = next_address.track
                buffer[NEXT_SECTOR_OFFSET] = next_address.sector
            else:
                buffer[NEXT_TRACK_OFFSET] = 0
                buffer[NEXT_SECTOR_OFFSET] = 0
            # Write up to 120 Track/Sector pairs
            for j in range(TRACK_SECTOR_OFFSET, SECTOR_SIZE, 2):
                if not data_sectors:
                    break
                data_address = data_sectors.pop(0)
                buffer[j] = data_address.track
                buffer[j + 1] = data_address.sector
            # Write the track/sector list sector
            self.fs.write_sector(buffer, trac_sector_list_sectors[i])
        # Search for an empty catalog entry
        entry = self.search_empty_entry()
        if entry is None:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        # Update the entry
        entry.is_deleted = False
        entry.is_locked = False
        entry.address = sectors[0]
        entry.filename = appledos_canonical_filename(fullname)  # type: ignore
        entry.raw_filename = appledos_filename_to_raw_filename(fullname)  # type: ignore
        entry.length = number_of_blocks + trac_sector_list_num
        entry.raw_file_type = appledos_get_raw_file_type(file_type)
        return entry

    def __str__(self) -> str:
        return dump_struct(self.__dict__, exclude=["fs", "directory_entries"])


class AppleDOSDirectoryEntry(AbstractDirectoryEntry):
    """
    Apple DOS 3.x Directory Entry

       Field                                Byte of
      Length                                Sector
             +----------------------------+
     1 byte  |    T/S List Sector track   | $00
             |----------------------------|
     1 byte  |   T/S List Sector sector   | $01
             +----------------------------+
     1 byte  |     file type and flag     | $02
             +----------------------------+
             |                            | $03
     30 byte |         filename           |
             |                            | $20
             +----------------------------+
     2 byte  |          file              | $21
             |         length             | $22
             +----------------------------+
    """

    fs: "AppleDOSFilesystem"
    address: TrackSector  # Track/Sector List address
    raw_file_type: int  # File type
    is_locked: bool  # Locked
    is_deleted: bool  # Deleted
    filename: str  # Filename
    raw_filename: bytes  # Raw filename
    length: int  # Length in sectors

    def __init__(self, fs: "AppleDOSFilesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "AppleDOSFilesystem", buffer: bytes, position: int) -> "AppleDOSDirectoryEntry":
        self = cls(fs)
        (
            track,  # track
            sector,  # sector
            raw_file_type,  # file type
            raw_filename,  # filename
            self.length,  # length
        ) = struct.unpack_from(FILE_DESCRIPTIVE_ENTRY_FORMAT, buffer, position)
        self.raw_file_type = raw_file_type & 0x7F
        self.is_locked = bool(raw_file_type & LOCKED_FLAG)
        if track == DELETED_TRACK:
            # The original track number is copied to the last byte of the filename
            track = raw_filename[-1]
            raw_filename = raw_filename[:-1]
            self.is_deleted = True
        else:
            self.is_deleted = False
        self.address = TrackSector(track, sector)
        self.raw_filename = raw_filename
        self.filename = "".join([chr(x & 0x7F) for x in raw_filename]).strip()
        return self

    @classmethod
    def create(cls, fs: "AppleDOSFilesystem") -> "AppleDOSDirectoryEntry":
        self = cls(fs)
        self.address = TrackSector(0, 0)
        self.raw_file_type = 0
        self.is_locked = False
        self.is_deleted = False
        self.filename = ""
        self.raw_filename = b""
        self.length = 0
        return self

    def write_buffer(self, buffer: bytearray, position: int = 0) -> None:
        """
        Write the directory entry
        """
        track = self.address.track
        sector = self.address.sector
        raw_filename = appledos_filename_to_raw_filename(self.filename)
        raw_file_type = self.raw_file_type | (LOCKED_FLAG if self.is_locked else 0)
        if self.is_deleted:
            track = DELETED_TRACK
            raw_filename = raw_filename[:-1] + bytes([self.address.track])
        struct.pack_into(
            FILE_DESCRIPTIVE_ENTRY_FORMAT,
            buffer,
            position,
            track,
            sector,
            raw_file_type,
            raw_filename,
            self.length,
        )

    @property
    def is_empty(self) -> bool:
        """
        If track is 0, the entry is assumed to never have been used and
        is available for use.
        """
        return self.address.track == 0

    @property
    def file_type(self) -> t.Optional[str]:
        """
        File type (e.g. DATA)
        """
        return FILE_TYPES.get(self.raw_file_type)

    def read_bytes(self, file_mode: t.Optional[str] = None) -> bytes:
        """Get the content of the file"""
        data = super().read_bytes()
        if self.file_type == "B":
            # Binary File Format on Disk
            # +------------------+-----------------+------------------
            # | Address (2 byte) | Length (2 byte) | Memory image ...
            # +------------------+-----------------+------------------
            # https://archive.org/details/beneath-apple-dos-prodos-2020/page/42/mode/2up
            address, length = struct.unpack_from(BINARY_FILE_FORMAT, data, 0)
            prodos_file_info = ProDOSFileInfo(0xFF, PRODOS_BIN_FILE_TYPE, address)
            data = encode_apple_single(prodos_file_info, data[struct.calcsize(BINARY_FILE_FORMAT) :])
        elif self.file_type == "A" or self.file_type == "I":
            # Integer/Applesoft Basic File Format on Disk
            # +-----------------+--------------------------
            # | Length (2 byte) | Program memory image ...
            # +-----------------+--------------------------
            # https://archive.org/details/beneath-apple-dos-prodos-2020/page/44/mode/2up
            length = struct.unpack_from(BASIC_FILE_FORMAT, data, 0)[0]
            data = data[struct.calcsize(BASIC_FILE_FORMAT) : -length]
        return data

    def blocks(self, include_indexes: bool = False) -> t.Iterator[TrackSector]:
        """
        Iterate over the sectors of the file

        Each file is associated with a "Track/Sector List"
        It contains a list of track/sector pointer pairs
        which sequentially list the data sectors which make up the file.

        https://archive.org/details/Beneath_Apple_DOS_alt/page/n21/mode/2up
        """
        address = self.address
        while address.track != 0:
            if include_indexes:
                yield address
            # Read the T/S list sector
            buffer = self.fs.read_sector(address)
            # Read up to 120 more Track/Sector pairs
            for i in range(TRACK_SECTOR_OFFSET, SECTOR_SIZE, 2):
                data_address = TrackSector(buffer[i], buffer[i + 1])
                if data_address.track == 0 and data_address.sector == 0:
                    break
                yield data_address
            # Track and sector number of next T/S List sector
            address = TrackSector(buffer[NEXT_TRACK_OFFSET], buffer[NEXT_SECTOR_OFFSET])

    @property
    def fullname(self) -> str:
        return self.filename

    @property
    def basename(self) -> str:
        return self.filename

    def get_length(self) -> int:
        """
        Get the length in sectors
        """
        return self.length

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.length * SECTOR_SIZE

    def get_block_size(self) -> int:
        """
        Get file sector size in bytes
        """
        return SECTOR_SIZE

    def delete(self) -> bool:
        """
        Delete the file
        """
        # Update the catalog
        found = False
        catalog = AppleDOSCatalog.read(self.fs)
        for entry in catalog.iterdir():
            if entry.address == self.address:
                entry.is_deleted = True
                found = True
                break
        if not found:
            return False
        catalog.write()
        # Update the bitmap
        vtoc = AppleDOSVTOC.read(self.fs)
        for address in self.blocks(include_indexes=True):
            if address.track != 0 or address.sector != 0:
                vtoc.set_free(address)
        vtoc.write()
        return True

    def write(self) -> bool:
        """
        Write the directory entry
        """
        catalog = AppleDOSCatalog.read(self.fs)
        for entry in catalog.iterdir():
            if entry.address == self.address:
                catalog.write()
                return True
        return False

    def open(self, file_mode: t.Optional[str] = None) -> AppleDOSFile:
        """
        Open a file
        """
        return AppleDOSFile(self, file_mode)

    def __str__(self) -> str:
        return (
            f"Track: {self.address.track:>3}  Sector: {self.address.sector:>3}  Length: {self.length:>5}  "
            f"File type: {self.file_type}  {'Deleted' if self.is_deleted else 'Locked' if self.is_locked else '':<7} "
            f"{self.filename}"
        )

    def __repr__(self) -> str:
        return str(self)


class AppleDOSFilesystem(AbstractAppleDiskFilesystem):
    """
    Apple II DOS 3.x Filesystem

    https://archive.org/details/Beneath_Apple_DOS_alt/page/n17/mode/2up
    https://archive.org/details/beneath-apple-dos-prodos-2020/page/30/mode/2up
    """

    fs_name = "appledos"
    fs_description = "Apple II DOS 3.x"
    fs_platforms = ["apple2"]

    catalog_address: TrackSector  # first catalog address
    number_of_tracks: int  # Number of tracks on disk
    sectors_per_track: int  # Sectors per track

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        **kwargs: t.Union[bool, str],
    ) -> "AppleDOSFilesystem":
        self = cls(file_or_dev)
        vtoc: t.Optional[AppleDOSVTOC] = None
        # Read VTOC in DOS order
        self.dev.prodos_order = False
        try:
            dos_vtoc = AppleDOSVTOC.read(self)
            dos_files = len(AppleDOSCatalog.read(self, dos_vtoc.catalog_address).directory_entries)
        except Exception:
            dos_vtoc = None
            dos_files = 0
        self.dev.prodos_order = True
        # Read VTOC in ProDOS order
        try:
            prodos_vtoc = AppleDOSVTOC.read(self)
            prodos_files = len(AppleDOSCatalog.read(self, prodos_vtoc.catalog_address).directory_entries)
        except Exception:
            prodos_vtoc = None
            prodos_files = 0
        # Choose the order with the most files
        if prodos_files > dos_files:
            self.dev.prodos_order = True
            vtoc: AppleDOSVTOC = prodos_vtoc  # type: ignore
        else:
            self.dev.prodos_order = False
            vtoc = dos_vtoc  # type: ignore
        if vtoc is None or not vtoc.is_valid:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        self.catalog_address = vtoc.catalog_address
        if vtoc.number_of_tracks > 0:
            self.number_of_tracks = vtoc.number_of_tracks
        if vtoc.sectors_per_track > 0:
            self.sectors_per_track = vtoc.sectors_per_track
        return self

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
    ) -> t.Iterator["AppleDOSDirectoryEntry"]:
        if pattern:
            pattern = appledos_canonical_filename(pattern, wildcard=wildcard)
        catalog = AppleDOSCatalog.read(self)
        for entry in catalog.iterdir(include_deleted=include_all):
            if filename_match(entry.basename, pattern, wildcard):
                yield entry

    @property
    def entries_list(self) -> t.Iterator["AppleDOSDirectoryEntry"]:
        catalog = AppleDOSCatalog.read(self)
        yield from catalog.iterdir()

    def get_file_entry(self, fullname: str, include_deleted: bool = False) -> AppleDOSDirectoryEntry:
        """
        Get the file entry for a given path
        """
        fullname = appledos_canonical_filename(fullname)  # type: ignore
        catalog = AppleDOSCatalog.read(self)
        for entry in catalog.iterdir(include_deleted=include_deleted):
            if entry.fullname == fullname:
                return entry
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        """
        Write content to a file
        """
        # Check if the file is an AppleSingle file and extract the content and metadata
        try:
            content, _, prodos_file_info = decode_apple_single(content)
            if prodos_file_info is not None and file_type is None:
                # Map ProDOS file type to DOS file type
                if prodos_file_info.file_type == PRODOS_TXT_FILE_TYPE:
                    file_type = "T"
                elif prodos_file_info.file_type == PRODOS_BIN_FILE_TYPE:
                    # Binary File Format on Disk
                    # +------------------+-----------------+------------------
                    # | Address (2 byte) | Length (2 byte) | Memory image ...
                    # +------------------+-----------------+------------------
                    # https://archive.org/details/beneath-apple-dos-prodos-2020/page/42/mode/2up
                    file_type = "B"
                    header = struct.pack(BINARY_FILE_FORMAT, prodos_file_info.aux_type, len(content))
                    content = header + content
                elif prodos_file_info.file_type in (PRODOS_INT_FILE_TYPE, PRODOS_BAS_FILE_TYPE):
                    # Integer/Applesoft Basic File Format on Disk
                    # +-----------------+--------------------------
                    # | Length (2 byte) | Program memory image ...
                    # +-----------------+--------------------------
                    # https://archive.org/details/beneath-apple-dos-prodos-2020/page/44/mode/2up
                    if prodos_file_info.file_type == PRODOS_INT_FILE_TYPE:
                        file_type = "I"
                    else:
                        file_type = "A"
                    header = struct.pack(BASIC_FILE_FORMAT, len(content))
                    content = header + content
                elif prodos_file_info.file_type == PRODOS_REL_FILE_TYPE:
                    file_type = "R"
        except ValueError:
            pass

        number_of_blocks = int(math.ceil(len(content) / SECTOR_SIZE))
        entry = self.create_file(
            fullname=fullname,
            number_of_blocks=number_of_blocks,
            creation_date=creation_date,
            file_type=file_type,
        )
        if entry is not None:
            content = content + (b"\0" * SECTOR_SIZE)  # pad with zeros
            f = entry.open(file_mode)
            try:
                f.write_block(content, block_number=0, number_of_blocks=number_of_blocks)
            finally:
                f.close()

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,  # optional file type
    ) -> t.Optional[AppleDOSDirectoryEntry]:
        """
        Create a new file
        """
        fullname = appledos_canonical_filename(fullname)  # type: ignore
        try:
            self.get_file_entry(fullname).delete()
        except FileNotFoundError:
            pass
        vtoc = AppleDOSVTOC.read(self)
        catalog = AppleDOSCatalog.read(self)
        entry = catalog.create_file(
            vtoc=vtoc,
            fullname=fullname,
            number_of_blocks=number_of_blocks,
            file_type=file_type,
        )
        vtoc.write()
        catalog.write()
        return entry

    def chdir(self, fullname: str) -> bool:
        return False

    def isdir(self, fullname: str) -> bool:
        return False

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        if not options.get("brief"):
            sys.stdout.write("DISK VOLUME\n\n")
        if pattern:
            pattern = appledos_canonical_filename(pattern, wildcard=True)
        for x in self.filter_entries_list(pattern, include_all=options.get("full", False)):
            if options.get("brief"):
                sys.stdout.write(f"{x.fullname}\n")
            else:
                prefix = "X" if x.is_deleted else "*" if x.is_locked else " "
                sys.stdout.write(f"{prefix}{x.file_type} {x.length:>03} {x.fullname}\n")

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        """
        Examine the filesystem
        """
        if options.get("bitmap"):
            # Examine the bitmap
            vtoc = AppleDOSVTOC.read(self)
            for track, bits in enumerate(vtoc.bitmaps):
                sys.stdout.write(f"Track {track:>3}: {bits:032b}\n")
        elif not arg:
            # Examine the entire filesystem
            vtoc = AppleDOSVTOC.read(self)
            sys.stdout.write(
                dump_struct(
                    vtoc.__dict__,
                    exclude=["fs", "bitmaps"],
                    include=["catalog_address"],
                )
            )
            sys.stdout.write("\n")
            catalog = AppleDOSCatalog.read(self)
            sys.stdout.write(
                dump_struct(
                    catalog.__dict__,
                    exclude=["fs", "directory_entries"],
                    include=["catalog_address"],
                )
            )
            sys.stdout.write("\n\nDirectory entries:\n")
            for i, entry in enumerate(catalog.directory_entries):
                if not entry.is_empty:
                    sys.stdout.write(f"{i:>3}#  {entry}\n")
        else:
            # Examine by path
            entry = self.get_file_entry(arg, include_deleted=True)
            entry_dict = dict(entry.__dict__)
            entry_dict["address"] = str(entry.address)
            entry_dict["blocks"] = list(entry.blocks())  # type: ignore
            sys.stdout.write(dump_struct(entry_dict, exclude=["fs"]) + "\n")
            vtoc = AppleDOSVTOC.read(self)
            for sector in entry.blocks(include_indexes=True):
                sys.stdout.write(f"Sector {sector} is {'free' if vtoc.is_free(sector) else 'used'}\n")

    def dump(self, fullname: t.Optional[str], start: t.Optional[int] = None, end: t.Optional[int] = None) -> None:
        """
        Dump the content of a file or a range of blocks
        """
        if fullname:
            if start is None:
                start = 0
            if end is None:
                entry = self.get_file_entry(fullname)
                end = entry.get_length() - 1
            f = self.open_file(fullname, file_mode=IMAGE)
            try:
                for block_number in range(start, end + 1):
                    data = f.read_block(block_number)
                    sys.stdout.write(f"\nBLOCK NUMBER   {block_number:08}\n")
                    if entry.raw_file_type == FILE_TYPE_TEXT:  # type: ignore
                        # Remove the high bit for text files
                        data = bytes([x & 0x7F for x in data])
                    hex_dump(data)
            finally:
                f.close()
        else:
            if start is None:
                start = 0
            if end is None:
                if start == 0:
                    end = self.number_of_tracks
                else:
                    end = start + 1
            for track in range(start, end):
                for sector in range(0, self.sectors_per_track):
                    data = self.read_sector(TrackSector(track, sector))
                    sys.stdout.write(f"\nTRACK {track:02}  SECTOR {sector:02}\n")
                    hex_dump(data)

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()

    @classmethod
    def initialize(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        **kwargs: t.Union[bool, str],
    ) -> "AppleDOSFilesystem":
        """
        Initialize the filesystem
        """
        self = cls(file_or_dev)
        vtoc = AppleDOSVTOC.create(self)
        catalog = AppleDOSCatalog.create(self, vtoc)
        catalog.write()
        vtoc.write()
        self.catalog_address = vtoc.catalog_address
        self.number_of_tracks = vtoc.number_of_tracks
        self.sectors_per_track = vtoc.sectors_per_track
        return self

    def get_pwd(self) -> str:
        return ""

    def get_types(self) -> t.List[str]:
        """
        Get the list of the supported file types
        """
        return list(FILE_TYPES.values())



# Copyright (C) 2414 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import os
import struct
import typing as t

from ..commons import ASCII, IMAGE, BlockDirection, Direction
from .block import BlockDevice

if t.TYPE_CHECKING:
    from ..abstract import AbstractFile

__all__ = [
    "BYTES_PER_WORD_18BIT",
    "WORDS_PER_BLOCK",
    "BlockDevice18Bit",
    "from_18bit_words_to_bytes",
    "from_bytes_to_18bit_words",
    "reverse_word",
]


BYTES_PER_WORD_18BIT = 4  # Each word is encoded in 4 bytes
WORDS_PER_BLOCK = 256  # Number of words per block


def from_18bit_words_to_bytes(words: list[int], file_type: str = ASCII) -> bytes:
    """
    Convert 18bit words to 3 bytes (IMAGE) or 2 bytes (ASCII)
    """
    data = bytearray()
    if file_type == ASCII:
        for word in words:
            data.append((word >> 9) & 0o177)
            data.append(word & 0o177)
    else:
        for word in words:
            data.append(((word >> 12) & 0o077) + 0x80)
            data.append(((word >> 6) & 0o077) + 0x80)
            data.append((word & 0o077) + 0x80)
    return bytes(data)


def from_bytes_to_18bit_words(data: bytes, file_type: str = ASCII) -> t.List[int]:
    """
    Convert 3 bytes to 18bit words, keeping only the lower 6 bits of each byte (IMAGE)
    or 2 bytes to 18bit words (ASCII)
    """
    words = []
    if file_type == ASCII:
        for i in range(0, len(data), 2):
            words.append((data[i] << 9) | data[i + 1])
    else:
        for i in range(0, len(data), 3):
            words.append(((data[i] - 0x80) << 12) | ((data[i + 1] - 0x80) << 6) | (data[i + 2] - 0x80))
    return words


def reverse_word(w: int) -> int:
    """
    Reverse the order of bits in a 18-bit word
    (used for backward reading/writing of DECtape blocks)
    """
    w = ~w
    return (
        ((w & 0o700000) >> 15)
        | ((w & 0o070000) >> 9)
        | ((w & 0o007000) >> 3)
        | ((w & 0o000700) << 3)
        | ((w & 0o000070) << 9)
        | ((w & 0o000007) << 15)
    )


class BlockDevice18Bit(BlockDevice):
    """
    Block device for 18-bit mode
    """

    words_per_block: int

    def __init__(
        self,
        file: "AbstractFile",
        words_per_block: int = WORDS_PER_BLOCK,
    ):
        super().__init__(file)
        self.words_per_block = words_per_block

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        data = bytearray()
        for i in range(block_number, block_number + number_of_blocks):
            words = self.read_words_block(block_number)
            data.extend(from_18bit_words_to_bytes(words, IMAGE))
        return bytes(data)

    def read_words_block(
        self,
        block: t.Union[int, BlockDirection],
    ) -> t.List[int]:
        """
        Read a 256 bytes block as 18bit words

        It is possible to read data from a DECtape in backward direction.
        However, a re-ordering of both the entire block and individual words is required.

        Pag 75
        https://bitsavers.org/pdf/dec/pdp15/DEC-15-H2DC-D_usersVol2.pdf
        """
        block_number: int = block.block_number if isinstance(block, BlockDirection) else block
        direction = block.direction if isinstance(block, BlockDirection) else Direction.FORWARD
        # Read the block
        self.f.seek(block_number * self.words_per_block * BYTES_PER_WORD_18BIT)
        buffer = self.f.read(self.words_per_block * BYTES_PER_WORD_18BIT)
        if len(buffer) < self.words_per_block * BYTES_PER_WORD_18BIT:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        words = list(struct.unpack(f"{self.words_per_block}I", buffer))
        # If the direction is backward, reverse the order of words and each word
        if direction == Direction.FORWARD:
            return words
        else:  # Direction.BACKWARD
            return [reverse_word(w) for w in reversed(words)]

    def write_words_block(
        self,
        block: t.Union[int, BlockDirection],
        words: t.List[int],
    ) -> None:
        """
        Write 256 18bit words as a block
        """
        block_number: int = block.block_number if isinstance(block, BlockDirection) else block
        direction = block.direction if isinstance(block, BlockDirection) else Direction.FORWARD
        # If the direction is backward, reverse the order of words and each word
        if direction == Direction.BACKWARD:
            words = [reverse_word(w) for w in reversed(words)]
        # Write the block
        self.f.seek(block_number * self.words_per_block * BYTES_PER_WORD_18BIT)
        buffer = struct.pack(f"{self.words_per_block}I", *words)
        self.f.write(buffer)



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import importlib.resources
import os
import shlex
import sys
import typing as t

from ..commons import ASCII, splitdrive
from ..volumes import DEFAULT_VOLUME, FILESYSTEMS
from .commons import (
    CommandError,
    Commands,
    ask,
    copy_file,
    extract_options,
    get_int_option,
)

if t.TYPE_CHECKING:
    from ..shell import Shell

__all__ = ["cmds"]


cmds = Commands()


@cmds.register("DIR_ECTORY", aliases=["LS"])
def directory(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
DIR             Lists file directories

  SYNTAX
        DIR [/options] [[volume:][filespec]]

  SEMANTICS
        This command generates a listing of the directory you specify.

  OPTIONS
   BRIEF
        Lists only file names and file types
   FULL
        Lists the entire directory, including unused areas
   UIC
        Lists all UIC on a device (DOS-11, RSTS/E)

  EXAMPLES
        DIR A:*.SAV
        DIR SY:

    """
    # fmt: on
    args, options = extract_options(args, "/brief", "/uic", "/full")
    if len(args) > 1:
        raise CommandError("?DIR-F-Too many arguments")
    if args:
        volume_id, pattern = splitdrive(args[0])
    else:
        volume_id = DEFAULT_VOLUME
        pattern = None
    fs = shell.volumes.get(volume_id, cmd="DIR")
    fs.dir(volume_id, pattern, options)  # type: ignore


@cmds.register("TY_PE")
def type(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
TYPE            Outputs files to the terminal

  SYNTAX
        TYPE [volume:]filespec

  EXAMPLES
        TYPE A.TXT

    """
    # fmt: on
    if not args:
        line = ask("File? ")
        args = shlex.split(line)
    if len(args) > 1:
        raise CommandError("?TYPE-F-Too many arguments")
    volume_id, pattern = splitdrive(args[0])
    fs = shell.volumes.get(volume_id, cmd="TYPE")
    match = False
    for entry in fs.filter_entries_list(pattern):
        match = True
        content = entry.read_bytes(file_mode=ASCII)
        if content is not None:
            os.write(sys.stdout.fileno(), content)
            sys.stdout.write("\n")
    if not match:
        raise CommandError("?TYPE-F-No files")


@cmds.register("COP_Y")
def copy(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
COPY            Copies files

  SYNTAX
        COPY [/options] [input-volume:]input-filespec [output-volume:][output-filespec]

  OPTIONS
   ASCII                Copy as ASCII text
   TYPE:type
        Specifies that the output file type, if supported by the target filesystem
        See the SHOW TYPES command for a list of filesystems.

  EXAMPLES
        COPY *.TXT DK:
        COPY /ASCII /TYPE:CONTIGUOUS LICENSE DK:

    """
    # fmt: on
    args, options = extract_options(args, "/ascii", "/type")
    if len(args) > 2:
        raise CommandError("?COPY-F-Too many arguments")
    file_mode = ASCII if options.get("ascii") else None
    cfrom = len(args) > 0 and args[0]
    to = len(args) > 1 and args[1]
    if not cfrom:
        cfrom = ask("From? ")
    from_volume_id, cfrom = splitdrive(cfrom)
    from_fs = shell.volumes.get(from_volume_id, cmd="COPY")
    if not to:
        to = ask("To? ")
    to_volume_id, to = splitdrive(to)
    to_fs = shell.volumes.get(to_volume_id, cmd="COPY")
    from_len = len(list(from_fs.filter_entries_list(cfrom)))
    from_list = from_fs.filter_entries_list(cfrom)
    file_type = options["type"].upper() if isinstance(options.get("type"), str) else None  # type: ignore
    if from_len == 0:  # No files
        raise CommandError("?COPY-F-No files")
    elif from_len == 1:  # One file to be copied
        source = list(from_list)[0]
        if not to:
            to_pwd = shell.volumes.get(to_volume_id).get_pwd()
            to_path = os.path.join(to_pwd, source.basename)
        elif to and to_fs.isdir(to):
            to_path = os.path.join(to, source.basename)
        else:
            to_path = to
        from_entry = from_fs.get_file_entry(source.fullname)
        if not from_entry:
            raise CommandError(f"?COPY-F-Error copying {source.fullname}")
        sys.stdout.write("%s:%s -> %s:%s\n" % (from_volume_id, source.fullname, to_volume_id, to_path))
        copy_file(from_fs, from_entry, to_fs, to_path, file_type, file_mode, shell.verbose, cmd="COPY")
    else:
        if not to:
            to = shell.volumes.get(to_volume_id).get_pwd()
        elif not to_fs.isdir(to):
            raise CommandError("?COPY-F-Target must be a volume or a directory")
        for from_entry in from_fs.filter_entries_list(cfrom):
            if to:
                to_path = os.path.join(to, from_entry.basename)
            else:
                to_path = from_entry.basename
            sys.stdout.write("%s:%s -> %s:%s\n" % (from_volume_id, from_entry.fullname, to_volume_id, to_path))
            copy_file(from_fs, from_entry, to_fs, to_path, file_type, file_mode, shell.verbose, cmd="COPY")


@cmds.register("DEL_ETE")
def delete(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
DELETE          Removes files from a volume

  SYNTAX
        DELETE [volume:]filespec

  SEMANTICS
        This command deletes the files you specify from the volume.

  EXAMPLES
        DELETE *.OBJ

    """
    # fmt: on
    if not args:
        line = ask("Files? ")
        args = shlex.split(line)
    volume_id, pattern = splitdrive(args[0])
    fs = shell.volumes.get(volume_id, cmd="DEL")
    match = False
    for x in fs.filter_entries_list(pattern, expand=False):  # don't expand directories
        match = True
        if not x.delete():
            sys.stdout.write("?DEL-F-Error deleting %s\n" % x.fullname)
    if not match:
        raise CommandError("?DEL-F-No files")


@cmds.register("E_XAMINE")
def examine(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
EXAMINE         Examines disk structure

  SYNTAX
        EXAMINE volume:

   FULL
        Lists the entire directory, including unused areas

    """
    # fmt: on
    args, options = extract_options(args, "/free", "/bitmap", "/diskid", "/full")
    if not args:
        args = ask("From? ").split()
    for arg in args:
        volume_id, fullname = splitdrive(arg)
        fs = shell.volumes.get(volume_id)
        fs.examine(fullname, options)


@cmds.register("DU_MP")
def dump(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
DUMP            Prints formatted data dumps of files or devices

  SYNTAX
        DUMP [/options] filespec

  SEMANTICS
        Filespec represents the device or file to be dumped.

  OPTIONS
   START:block
        Specifies the first block to be dumped
   END:block
        Specifies the last block to be dumped

  EXAMPLES
        DUMP A.OBJ
        DUMP /START:6 /END:6 DL0:

    """
    # fmt: on
    args, options = extract_options(args, "/start", "/end")
    start = get_int_option(options, "start")
    end = get_int_option(options, "end")
    if not args:
        args = ask("From? ").split()
    for arg in args:
        try:
            volume_id, fullname = splitdrive(arg)
            fs = shell.volumes.get(volume_id)
            fs.dump(fullname, start=start, end=end)
        except FileNotFoundError:
            raise CommandError("?DUMP-F-File not found")


@cmds.register("CR_EATE")
def create(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
CREATE          Creates files or directories

  SYNTAX
        CREATE [/options] [volume:]filespec

  SEMANTICS
        This command creates a file or directory on the volume you specify.
        The default option is to create a file.

  OPTIONS
   FILE
        Creates a file with a specific name and size
   DIRECTORY
        Creates a directory
   ALLOCATE:size
        Specifies the number of blocks to allocate to the created file
   TYPE:type
        Specifies the file type
        See the SHOW TYPES command for a list of filesystems.

  EXAMPLES
        CREATE NEW.DSK /ALLOCATE:200

    """
    # fmt: on
    args, options = extract_options(args, "/file", "/dir", "/directory", "/uic", "/allocate", "/type")
    if len(args) > 1:
        raise CommandError("?CREATE-F-Too many arguments")
    if "file" in options:
        kind = "file"
    if "directory" in options or "dir" in options:
        kind = "directory"
    else:
        kind = "file"
    path = len(args) > 0 and args[0]
    if not path:
        path = ask("File? ")
    volume_id, fullname = splitdrive(path)
    fs = shell.volumes.get(volume_id, cmd="CREATE")
    if kind == "directory":
        # Create a directory
        fs.create_directory(fullname, options)
    else:
        # Create a file
        allocate = options.get("allocate")
        if not allocate:
            allocate = ask("Size? ")
        try:
            number_of_blocks = int(allocate)
            if number_of_blocks < 0:
                raise ValueError
        except:
            raise CommandError("?CREATE-F-Invalid value specified with option")
        fs.create_file(
            fullname,
            number_of_blocks=number_of_blocks,
            file_type=options.get("type") if isinstance(options.get("type"), str) else None,  # type: ignore
        )


@cmds.register("MO_UNT")
def mount(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
MOUNT           Assigns a logical disk unit to a file

  SYNTAX
        MOUNT [/options] volume: [volume:]filespec

  SEMANTICS
        Associates a logical disk unit with a file.
        See the SHOW FILESYSTEMS command for a list of filesystems
        that can be mounted.

  EXAMPLES
        MOUNT AB: SY:rt11v503.dsk
        MOUNT /DOS11 AB: SY:dos.dsk
        MOUNT /UNIX7 AB: SY:unix7.dsk
        MOUNT /DMS /DEV:DECTAPE AB: SY:dms.tu56

  OPTIONS
   DEV:device type
        Specifies the device type, if supported by the target filesystem.

    """
    # fmt: on
    fs_args = [f"/{x}" for x in FILESYSTEMS.keys()]
    args, options = extract_options(args, "/dev", *fs_args)
    try:
        device_type: t.Optional[str] = options["dev"].lower()  # type: ignore
    except Exception:
        device_type = None
    if len(args) > 2:
        raise CommandError("?MOUNT-F-Too many arguments")
    logical = len(args) > 0 and args[0]
    path = len(args) > 1 and args[1]
    if not logical:
        logical = ask("Volume? ")
    if not path:
        path = ask("File? ")
    fstype = None
    for filesystem in FILESYSTEMS.keys():
        if options.get(filesystem):
            fstype = filesystem
            break
    shell.volumes.mount(path, logical, fstype=fstype, verbose=shell.verbose, device_type=device_type)


@cmds.register("DIS_MOUNT")
def dismount(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
DISMOUNT        Disassociates a logical disk assignment from a file

  SYNTAX
        DISMOUNT logical_name

  SEMANTICS
        Removes the association of a logical disk unit with its currently
        assigned file, thereby freeing it to be assigned to another file.

  EXAMPLES
        DISMOUNT AB:

    """
    # fmt: on
    if len(args) > 1:
        raise CommandError("?DISMOUNT-F-Too many arguments")
    if args:
        logical = args[0]
    else:
        logical = ask("Volume? ")
    shell.volumes.dismount(logical)


@cmds.register("AS_SIGN")
def assign(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
ASSIGN          Associates a logical device name with a device

  SYNTAX
        ASSIGN device-name logical-device-name

  SEMANTICS
        Associates a logical device name with a device.
        Logical-device-name is one to three alphanumeric characters long.

  EXAMPLES
        ASSIGN DL0: INP:

    """
    # fmt: on
    args, options = extract_options(args)
    if len(args) > 2:
        raise CommandError("?ASSIGN-F-Too many arguments")
    volume_id = len(args) > 0 and args[0]
    logical = len(args) > 1 and args[1]
    if not volume_id:
        volume_id = ask("Device name? ")
    if not logical:
        logical = ask("Logical name? ")
    shell.volumes.assign(volume_id, logical, verbose=shell.verbose)


@cmds.register("DEA_SSIGN")
def deassign(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
DEASSIGN        Removes logical device name assignments

  SYNTAX
        DEASSIGN logical-device-name

  SEMANTICS
        The DEASSIGN command disassociates a logical name.

  EXAMPLES
        DEASSIGN INP:

    """
    # fmt: on
    if len(args) > 1:
        raise CommandError("?DEASSIGN-F-Too many arguments")
    if args:
        logical = args[0]
    else:
        logical = ask("Volume? ")
    shell.volumes.deassign(logical, cmd="DEASSIGN")


@cmds.register("INI_TIALIZE")
def initialize(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
INITIALIZE      Writes an empty device directory on the specified volume

  SYNTAX
        INITIALIZE [/options] [volume:][filespec]]

  SEMANTICS
        Initializes the specified filesystem on the volume.
        Any data on the volume is lost.
        See the SHOW FILESYSTEMS command for a list of filesystems.

  EXAMPLES
        INIT /DOS11 /DEV:dectape test.tap

  OPTIONS
   NAME:name
        Specifies the volume name
   DEV:device type
        Specifies the device type, if supported by the target filesystem.

    """
    # fmt: on
    fs_args = [f"/{x}" for x in FILESYSTEMS.keys()]
    args, options = extract_options(args, "/name", "/dev", *fs_args)
    if len(args) > 1:
        raise CommandError("?INITIALIZE-F-Too many arguments")
    target = len(args) > 0 and args[0]
    if not target:
        target = ask("Volume? ")
    if target.endswith(":"):
        try:
            options["device_type"] = options["dev"].lower()  # type: ignore
        except Exception:
            pass
        fs = shell.volumes.get(target)
        fs.initialize(fs.dev, **options)
    else:
        filesystem_cls = None
        for k, v in FILESYSTEMS.items():
            if options.get(k):
                filesystem_cls = v
                break
        if filesystem_cls is None:
            raise CommandError("?INITIALIZE-F-Filesystem not specified")
        parent_volume_id, target_path = splitdrive(target)
        parent_fs = shell.volumes.get(parent_volume_id)
        target_file = parent_fs.open_file(target_path)
        try:
            device_type = options["dev"].lower()  # type: ignore
        except Exception:
            device_type = shell.volumes.guess_device_type(target_path)
        if device_type:
            options["device_type"] = device_type
        fs = filesystem_cls.initialize(target_file, **options)
        fs.close()


@cmds.register("CD")
def cd(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
CD              Changes or displays the current working drive and directory

  SYNTAX
        CD [[volume:][filespec]]

    """
    # fmt: on
    if len(args) > 1:
        raise CommandError("?CD-F-Too many arguments")
    elif len(args) == 0:
        sys.stdout.write("%s\n" % shell.volumes.get_pwd())
    elif not shell.volumes.chdir(args[0]):
        raise CommandError("?CD-F-Directory not found")


@cmds.register("@", aliases=["BATCH"])
def batch(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
@               Executes a command file

  SYNTAX
        @[volume:]filespec

  SEMANTICS
        You can group a collection of commands that you want to execute
        sequentially into a command file.
        This command executes the command file.

  EXAMPLES
        @MAKE.COM

    """
    # fmt: on
    if not args:
        return
    try:
        volume_id, filespec = splitdrive(args[0])
        fs = shell.volumes.get(volume_id, cmd="BATCH")
        for line in fs.read_text(filespec).split("\n"):
            if line.startswith("!"):
                continue
            shell.onecmd(line.strip(), catch_exceptions=False, batch=True)
    except SystemExit:
        pass
    except FileNotFoundError:
        raise CommandError("?KMON-F-File not found")


@cmds.register("PWD")
def pwd(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
PWD             Displays the current working drive and directory

  SYNTAX
        PWD

    """
    sys.stdout.write("%s\n" % shell.volumes.get_pwd())


@cmds.register("SH_OW")
def show(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
SHOW            Displays software status

  SYNTAX
        SHOW [options] [volume:]

  SEMANTICS
	    SHOW displays the device assignments; other information
	    is displayed by specifying one or more option names.

  OPTIONS
   FILESYSTEMS
        Show the supported filesystems
   TYPES
        Show the file types of a volume
   VERSION
        Show the version of XFERX
   VOLUMES
        Show the device assignments

  EXAMPLES
        SHOW
        SHOW FILESYSTEMS
        SHOW TYPES DL0:

    """
    # fmt: on
    func = cmds.get(f"SHOW {args[0]}" if args else "SHOW VOLUMES")
    if func is None:
        raise CommandError("?SHOW-F-Illegal command")
    func(shell, args)


@cmds.register("SHOW T_YPES")
def show_type(shell: "Shell", args: t.List[str]) -> None:
    """
    SHOW TYPES        Show the file types of a volume
    """
    # fmt: on
    if len(args) == 1:
        volume_id = ask("Volume? ")
    else:
        volume_id = args[1]
    fs = shell.volumes.get(volume_id)
    sys.stdout.write("File Types\n")
    sys.stdout.write("----------\n")
    for item in fs.get_types():
        sys.stdout.write(f"{item}\n")


@cmds.register("SHOW F_ILESYSTEMS", aliases=["SHOW FS"])
def show_filesystems(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
SHOW FILESYSTEMS  Show the supported filesystems
    """
    # fmt: on
    sys.stdout.write("Filesystems\n")
    sys.stdout.write("-----------\n")
    for k, v in sorted(FILESYSTEMS.items()):  # type: ignore
        sys.stdout.write(f"{k.upper():<10} {v.fs_description}\n")


@cmds.register("SHOW VE_RSION")
def show_version(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
SHOW VERSION      Show the version of XFERX
    """
    # fmt: on
    with importlib.resources.files("xferx").joinpath("VERSION").open("r", encoding="utf-8") as f:
        version = f.read().strip()
    sys.stdout.write(f"XFERX {version}\n")


@cmds.register("SHOW VO_LUMES")
def show_volumes(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
SHOW VOLUMES      Show the device assignments
    """
    # fmt: on
    sys.stdout.write("Volumes\n")
    sys.stdout.write("-------\n")
    for k, v in shell.volumes.volumes.items():
        label = f"{k}:"
        sys.stdout.write(f"{label:<6} {v}\n")
    for k, v in shell.volumes.logical.items():  # type: ignore
        label = f"{k}:"
        sys.stdout.write(f"{label:<4} = {v}:\n")


@cmds.register("EXIT", aliases=["QUIT"])
def exit(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
EXIT            Exit the shell

  SYNTAX
        EXIT
    """
    # fmt: on
    raise SystemExit


@cmds.register("H_ELP")
def help(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
HELP            Displays commands help

  SYNTAX
        HELP [topic]

    """
    # fmt: on
    if args and args[0] != "*":
        arg = args[0]
        func = cmds.get(arg)
        if func is None:
            raise CommandError(f"?HELP-F-Help not available for {arg}")
        sys.stdout.write(f"{func.__doc__}\n")
    else:
        for name, command in sorted(cmds.commands.items()):
            if not " " in name and command.__doc__ is not None:
                sys.stdout.write(command.__doc__.split("\n")[1])
                sys.stdout.write("\n")


@cmds.register("SHELL")
def shell(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
SHELL           Executes a system shell command

  SYNTAX
        SHELL command

    """
    # fmt: on
    os.system(shlex.join(args))


@cmds.register("ECHO")
def echo(shell: "Shell", args: t.List[str]) -> None:
    # fmt: off
    """
ECHO            Write arguments to the terminal

  SYNTAX
        ECHO [arg ...]

  EXAMPLES
        ECHO Hello World

    """
    # fmt: on
    line = " ".join(args).strip()
    sys.stdout.write(f"{line}\n")



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__all__ = [
    "RAD50",
    "rad50_word_to_asc",
    "rad2asc",
    "asc_to_rad50_word",
    "asc2rad",
]

from ..commons import bytes_to_word, word_to_bytes

RAD50_ALT = "\0ABCDEFGHIJKLMNOPQRSTUVWXYZ$.%0123456789:"
RAD50 = "\0ABCDEFGHIJKLMNOPQRSTUVWXYZ$%*0123456789:"


def rad50_word_to_asc(val: int) -> str:
    """
    Convert RAD50 word to 0-3 chars of ASCII
    """
    # split out RAD50 digits into three ASCII characters a/b/c
    c = RAD50[val % 0x28]
    b = RAD50[(val // 0x28) % 0x28]
    a = RAD50[val // (0x28 * 0x28)]
    result = ""
    if a != "\0":
        result += a
    if b != "\0":
        result += b
    if c != "\0":
        result += c
    return result


def rad2asc(buffer: bytes, position: int = 0) -> str:
    """
    Convert RAD50 2 bytes to 0-3 chars of ASCII
    """
    return rad50_word_to_asc(bytes_to_word(buffer, position=position))


def asc_to_rad50_word(val: str) -> int:
    """
    Convert a string of 3 ASCII to a RAD50 word
    """
    val1 = [RAD50.find(c.upper()) for c in val] + [0, 0, 0]
    val2 = [x > 0 and x or 0 for x in val1]
    return (val2[0] * 0x28 + val2[1]) * 0x28 + val2[2]


def asc2rad(val: str) -> bytes:
    """
    Convert a string of 3 ASCII to a RAD50 2 bytes
    """
    return word_to_bytes(asc_to_rad50_word(val))



import pytest

from xferx.pdp11.solofs import SOLOBitmap, SOLOFilesystem
from xferx.shell import Shell

DSK = "tests/dsk/solo.dsk"


def test_solo_read():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /solo {DSK}", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, SOLOFilesystem)

    shell.onecmd("dir t:[*,*]", batch=True)
    shell.onecmd("type t:1.txt", batch=True)

    x = fs.read_bytes("50.txt")
    x = x.rstrip(b"\0")
    assert len(x) == 2200
    for i in range(0, 50):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x

    l = list(fs.entries_list)
    filenames = [x.filename for x in l if not x.is_empty]
    assert "NEXT" in filenames

    entry = fs.get_file_entry("NEXT")
    assert entry.protected


def test_solo_write():
    shell = Shell(verbose=True)
    shell.onecmd(f"copy {DSK} {DSK}.mo", batch=True)
    shell.onecmd(f"mount in: /solo {DSK}", batch=True)
    shell.onecmd(f"mount ou: /solo {DSK}.mo", batch=True)
    fs = shell.volumes.get('OU')
    assert isinstance(fs, SOLOFilesystem)

    d = fs.get_file_entry("50.TXT")
    assert d is not None

    # Delete a file
    d.delete()
    with pytest.raises(FileNotFoundError):
        fs.get_file_entry("50.TXT")

    # Create a file
    shell.onecmd("copy in:10.TXT ou:10NEW.TXT", batch=True)
    x2 = fs.read_bytes("10NEW.txt")
    x2 = x2.rstrip(b"\0")
    assert len(x2) == 440
    for i in range(0, 10):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x2


def test_solo_init_write():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount in: /solo {DSK}", batch=True)
    shell.onecmd(f"create /allocate:4800 {DSK}.mo", batch=True)
    shell.onecmd(f"init /solo {DSK}.mo", batch=True)
    shell.onecmd(f"mount ou: /solo {DSK}.mo", batch=True)
    shell.onecmd("dir ou:", batch=True)
    shell.onecmd("copy in:*.TXT ou:", batch=True)
    fs = shell.volumes.get('OU')

    x = fs.read_bytes("50.txt")
    x = x.rstrip(b"\0")
    assert len(x) == 2200
    for i in range(0, 50):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x

    entry = fs.get_file_entry("50.txt")
    hash_key = entry.hash_key
    assert fs.get_searchlength(hash_key) == 1

    # Test init mounted volume
    shell.onecmd("init ou:", batch=True)
    with pytest.raises(Exception):
        fs.read_bytes("50.txt")


def test_dos11_bitmap():
    shell = Shell(verbose=True)
    shell.onecmd(f"copy {DSK} {DSK}.mo", batch=True)
    shell.onecmd(f"mount in: /solo {DSK}", batch=True)
    shell.onecmd(f"mount ou: /solo {DSK}.mo", batch=True)
    fs = shell.volumes.get('OU')
    assert isinstance(fs, SOLOFilesystem)

    # Test is_free
    entry = fs.get_file_entry("50.txt")
    bitmap = SOLOBitmap.read(fs)
    free = bitmap.free()
    page_map = list(entry.page_map)
    for i in page_map:
        assert not bitmap.is_free(i)
    length = entry.length
    entry.delete()
    bitmap = SOLOBitmap.read(fs)
    for i in page_map:
        assert bitmap.is_free(i)
    assert bitmap.free() == free + length + 1

    bitmap.allocate(10)
    f0 = bitmap.find_first_free()
    f1 = bitmap.find_first_free()
    assert f0 == f1
    bitmap.write()
    f2 = bitmap.find_first_free()
    print(f2)
    bitmap.write()


def test_solo_segments():
    shell = Shell(verbose=True)
    shell.onecmd(f"copy {DSK} {DSK}.mo", batch=True)
    shell.onecmd(f"mount in: /solo {DSK}", batch=True)
    shell.onecmd(f"mount ou: /solo {DSK}.mo", batch=True)
    fs = shell.volumes.get('OU')
    assert isinstance(fs, SOLOFilesystem)

    # Get segment entries
    entry = fs.get_file_entry("@KERNEL")
    assert entry.length == 24
    assert entry.protected
    assert entry.file_type == "SEGMENT"
    with pytest.raises(Exception):
        assert entry.delete()
    entry = fs.get_file_entry("@SOLO")
    assert entry.length == 64
    entry = fs.get_file_entry("@OTHEROS")
    assert entry.length == 64
    with pytest.raises(Exception):
        fs.get_file_entry("@NOTFOUND")

    # Create a segment
    shell.onecmd("copy /type:concode in:50.TXT ou:@SOLO", batch=True)
    x2 = fs.read_bytes("@SOLO")
    x2 = x2.rstrip(b"\0")
    assert len(x2) > 2000
    for i in range(0, 50):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x2



import pytest

from xferx.native import NativeFilesystem
from xferx.pdp11.rt11fs import RT11Filesystem
from xferx.shell import Shell


def test_help():
    shell = Shell(verbose=True)
    # Help
    shell.onecmd("HELP", batch=True)
    shell.onecmd("HELP HELP", batch=True)
    shell.onecmd("HELP *", batch=True)


def test_assign():
    shell = Shell(verbose=True)
    # Assign/Deassign
    with pytest.raises(Exception):
        shell.onecmd("ASSIGN XX: T1:", batch=True)
    shell.onecmd("ASSIGN SY: T1:", batch=True)
    shell.onecmd("ASSIGN SY: T2:", batch=True)
    shell.onecmd("ASSIGN T2: T3:", batch=True)
    shell.onecmd("DIR /BRIEF T3:", batch=True)
    shell.onecmd("DEASSIGN T3:", batch=True)
    with pytest.raises(Exception):
        shell.onecmd("DIR /BRIEF T3:", batch=True)
    with pytest.raises(Exception):
        shell.onecmd("DEASSIGN T3:", batch=True)


def test_cmds():
    shell = Shell(verbose=True)
    sys_fs = shell.volumes.get('DK')
    license = sys_fs.read_bytes("LICENSE")
    with pytest.raises(FileNotFoundError):
        sys_fs.read_bytes("not found")
    assert license
    assert isinstance(sys_fs, NativeFilesystem)
    assert len(list(sys_fs.entries_list)) > 0
    # Open file
    with pytest.raises(FileNotFoundError):
        sys_fs.open_file("LI")
    f = sys_fs.open_file("LICENSE")
    t = f.read_block(block_number=0, number_of_blocks=1000)
    assert f.size == len(t)
    f.close()
    # Create ad initialize the RT-11 filesystem
    shell.onecmd("CREATE test0.dsk /allocate:500", batch=True)
    shell.onecmd("INITIALIZE /RT11 test0.dsk", batch=True)
    shell.onecmd("MOUNT T: test0.dsk", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, RT11Filesystem)
    with pytest.raises(FileNotFoundError):
        fs.read_bytes("not found")
    l0 = len(list(fs.entries_list))
    # Copy a file
    shell.onecmd("COPY LICENSE T:", batch=True)
    assert len(list(fs.entries_list)) == l0 + 1
    license_rt11 = fs.read_bytes("LICENSE").rstrip(b'\0')
    assert license == license_rt11
    # Open file
    with pytest.raises(FileNotFoundError):
        fs.open_file("LI")
    f = fs.open_file("LICENSE")
    t = f.read_block(block_number=0, number_of_blocks=1000)
    assert f.size == len(t)
    f.close()
    # Create a disk in the RT11 fs
    shell.onecmd("T:", batch=True)
    shell.onecmd("EXAMINE T:", batch=True)
    with pytest.raises(OSError):
        shell.onecmd("CREATE /file test1.dsk /allocate:1000", batch=True)
    shell.onecmd("CREATE /file test1.dsk /allocate:100", batch=True)
    shell.onecmd("MOUNT T1: test1.dsk", batch=True)
    shell.onecmd("EXAMINE T1:", batch=True)
    shell.onecmd("INITIALIZE T1:", batch=True)
    shell.onecmd("DIR T1:*.*", batch=True)
    shell.onecmd("COPY T:L*.* T1:", batch=True)
    shell.onecmd("DISMOUNT T1:", batch=True)
    shell.onecmd("T:", batch=True)
    shell.onecmd("DEL test1.dsk", batch=True)
    # Misc commands
    shell.onecmd("PWD", batch=True)
    shell.onecmd("SHOW", batch=True)
    shell.onecmd("DIR", batch=True)
    shell.onecmd("EXAMINE T:", batch=True)
    shell.onecmd("EXAMINE T:LICENSE", batch=True)
    shell.onecmd("DUMP /start:6 /end:6 T:", batch=True)
    shell.onecmd("TYPE T:LICENSE", batch=True)
    # Delete the file
    shell.onecmd("DEL LICENSE", batch=True)
    assert len(list(fs.entries_list)) == l0
    shell.onecmd("SY:", batch=True)
    # Copy/delete multiple files
    shell.onecmd("COPY py* T:", batch=True)
    shell.onecmd("DELETE T:py*", batch=True)
    # Delete the test disk
    shell.onecmd("DISMOUNT T:", batch=True)
    shell.onecmd("DEL test0.dsk", batch=True)



# Copyright (C) 2414 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import struct
import typing as t

from ..abstract import AbstractFile
from ..commons import BLOCK_SIZE, swap_words
from ..device.abstract import AbstractDevice
from .commons import UNIXFilesystem, UNIXInode, format_mode, iterate_words

__all__ = ["UNIX4Filesystem", "UNIXInode4"]

# Version 4 - 6

V4_USED = 0o100000  # i-node is allocated
V4_BLK = 0o060000  # block device
V4_DIR = 0o040000  # directory
V4_CHR = 0o020000  # character device
V4_LARGE = 0o010000  # large file

V4_SUID = 0o4000  # set user ID on execution
V4_SGID = 0o2000  # set group ID on execution
V4_STXT = 0o1000  # sticky bit

V4_ROWN = 0o400  # read by owner
V4_WOWN = 0o200  # write by owner
V4_XOWN = 0o100  # execute by owner
V4_RGRP = 0o040  # read by group
V4_WGRP = 0o020  # write by group
V4_XGRP = 0o010  # execute by group
V4_ROTH = 0o004  # read by other
V4_WOTH = 0o002  # write by other
V4_XOTH = 0o001  # execute by other

V4_PERMS = [
    [(V4_BLK, "b"), (V4_DIR, "d"), (V4_CHR, "c"), (0, "-")],
    [(V4_ROWN, "r"), (0, "-")],
    [(V4_WOWN, "w"), (0, "-")],
    [(V4_SUID, "s"), (V4_XOWN, "x"), (0, "-")],
    [(V4_RGRP, "r"), (0, "-")],
    [(V4_WGRP, "w"), (0, "-")],
    [(V4_SGID, "s"), (V4_XGRP, "x"), (0, "-")],
    [(V4_ROTH, "r"), (0, "-")],
    [(V4_WOTH, "w"), (0, "-")],
    [(V4_XOTH, "x"), (0, "-")],
    [(V4_STXT, "t"), (0, " ")],
]

V4_NICFREE = 100  # number of superblock free blocks
V4_NICINOD = 100  # number of superblock inodes
V4_SUPER_BLOCK = 1  # Superblock
V4_SUPER_BLOCK_FORMAT = f"<HHH {V4_NICFREE}H H {V4_NICINOD}H BBB L"

V4_INODE_FORMAT = "<HBBBBH 16s II"
V4_FILENAME_LEN = 14
V4_INODE_SIZE = 32
V4_NADDR = 8
V4_DIR_FORMAT = f"H{V4_FILENAME_LEN}s"
V4_ROOT_INODE = 1
assert struct.calcsize(V4_INODE_FORMAT) == V4_INODE_SIZE


class UNIXInode4(UNIXInode):

    @classmethod
    def read(cls, fs: "UNIXFilesystem", inode_num: int, buffer: bytes, position: int = 0) -> "UNIXInode":
        self = UNIXInode4(fs)
        self.inode_num = inode_num
        (
            self.flags,  #   1 word  flags
            self.nlinks,  #  1 byte  number of links to file
            self.uid,  #     1 byte  user ID of owner
            self.gid,  #     1 byte  group ID of owner
            sz0,  #          1 byte  high byte of 24-bit size
            sz1,  #          1 word  low word of 24-bit size
            addr,  #         8 words block numbers or device numbers
            self.atime,  #   1 long  time of last access
            self.mtime,  #   1 long  time of last modification
        ) = struct.unpack_from(V4_INODE_FORMAT, buffer, position)
        self.addr = struct.unpack_from(f"{V4_NADDR}H", addr)  # type: ignore
        self.size = (sz0 << 16) + sz1
        self.atime = swap_words(self.atime)
        self.mtime = swap_words(self.mtime)
        return self

    def blocks(self) -> t.Iterator[int]:
        if self.is_large:
            # Large file
            for block_number in self.addr:
                if block_number == 0:
                    break
                indirect_block = self.fs.read_block(block_number)
                for n in iterate_words(indirect_block):
                    if n == 0:
                        break
                    yield n
        else:
            # Small file
            for block_number in self.addr:
                if block_number == 0:
                    break
                yield block_number

    @property
    def isdir(self) -> bool:
        return (self.flags & V4_DIR) == V4_DIR

    @property
    def is_regular_file(self) -> bool:
        return not self.isdir

    @property
    def is_large(self) -> bool:
        return bool(self.flags & V4_LARGE)

    @property
    def is_allocated(self) -> bool:
        return bool(self.flags & V4_USED)

    def __str__(self) -> str:
        if not self.is_allocated:
            return f"{self.inode_num:>4}# ---"
        else:
            mode = format_mode(self.flags, version=self.fs.version, perms=self.fs.perms)
            return f"{self.inode_num:>4}# {self.uid:>3},{self.gid:<3} nlinks: {self.nlinks:>3} size: {self.size:>8}  {mode} flags: {self.flags:06o}"


class UNIX4Filesystem(UNIXFilesystem):
    """
    UNIX version 4, 5, 6 Filesystem
    """

    fs_name = "unix4"
    fs_description = "UNIX version 4"
    fs_platforms = ["pdp-11"]
    version: int = 4  # UNIX version
    inode_size = V4_INODE_SIZE
    dir_format = V4_DIR_FORMAT
    root_inode = V4_ROOT_INODE
    perms = V4_PERMS
    unix_inode_class = UNIXInode4

    @classmethod
    def mount(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "UNIX4Filesystem":
        self = cls(file_or_dev)
        self.pwd = "/"
        self.read_superblock()
        return self

    def read_superblock(self) -> None:
        """Read superblock"""
        superblock_data = self.read_block(V4_SUPER_BLOCK)
        superblock = struct.unpack_from(V4_SUPER_BLOCK_FORMAT, superblock_data, 0)
        self.inode_list_blocks = superblock[0]
        self.volume_size = superblock[1]
        self.free_blocks_in_list = superblock[2]
        self.free_blocks_list = list(superblock[3 : 3 + V4_NICFREE])
        self.free_inodes_in_list = superblock[3 + V4_NICFREE]
        self.free_inodes_list = list(superblock[4 + V4_NICFREE : 4 + V4_NICFREE + V4_NICINOD])
        # _ = superblock[4 + V4_NICINOD + V4_NICFREE]  # lock during free list manipulation
        # _ = superblock[5 + V4_NICINOD + V4_NICFREE]  # lock during i-list manipulation
        # _ = superblock[6 + V4_NICINOD + V4_NICFREE]  # flag to indicate that the super-block has changed and should be written
        # _ = swap_words(superblock[7 + V4_NICINOD + V4_NICFREE]) # last super block update
        self.inodes = (self.inode_list_blocks - 1) * (BLOCK_SIZE // self.inode_size)



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY FILE_TYPES, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import math
import os
import struct
import sys
import typing as t
from datetime import date

from ..abstract import AbstractDirectoryEntry, AbstractFile
from ..commons import ASCII, BLOCK_SIZE, IMAGE, READ_FILE_FULL, filename_match
from ..device.abstract import AbstractDevice
from .abstract import AbstractRXBlockFilesystem

__all__ = [
    "SOLOFile",
    "SOLODirectoryEntry",
    "SOLOFilesystem",
    "solo_canonical_filename",
]

EM = b"\x19"  # End of Medium

DISK_SIZE = 4800  # Disk size in blocks
ID_LENGTH = 12  # Filename length
ENTRY_FORMAT = "<12sHHH10sHH"  # Catalog entry format
ENTRY_LENGTH = 32  # Catalog entry length
CAT_PAGE_LENGTH = BLOCK_SIZE // ENTRY_LENGTH  # Entries in a catalog page
BLOCKS_PER_CYLINDER = 24  # Blocks per cylinder
CYLINDERS_PER_GROUP = 5  # Cylinders per group
GROUP_LENGTH = BLOCKS_PER_CYLINDER * CYLINDERS_PER_GROUP  # Blocks per group (120)
KERNEL_LENGTH = 24  # Kernel length (in blocks)
SEGMENT_LENGTH = 64  # Segment length (in blocks)
FREE_LIST_LENGTH = 2  # Free blocks bitmap length in blocks
# Bitset 0 - 119 => 120 bits => 15 chars, 7.5 words (8 words, 1 unused byte)
FREE_PAGE_GROUP_LENGTH = GROUP_LENGTH // 8  # 15 bytes
FREE_PAGE_GROUP_PAD = 1  # 1byte padding
FREE_PAGE_LENGTH = 31  # number of groups in a free page
FREE_PAGE_MISC_FORMAT = "<2H12x"  # Misc entry format
FREE_PAGE_MISC_LENGTH = 16
MAX_FILE_SIZE = 255  # Max file length in blocks

KERNEL_ADDR = 0  # Kernel block number
SOLO_OS_ADDR = KERNEL_ADDR + KERNEL_LENGTH  # 24 Solo OS segment block number
OTHER_OS_ADDR = SOLO_OS_ADDR + SEGMENT_LENGTH  # 88 Other OS segment block number
FREE_LIST_ADDR = OTHER_OS_ADDR + SEGMENT_LENGTH  # 152 Free blocks bitmap block number
CAT_ADDR = FREE_LIST_ADDR + FREE_LIST_LENGTH  # 154 - Catalog block number

# Segments are 3 contiguous, fixed-size files
SEGMENTS = {
    KERNEL_ADDR: "@KERNEL",
    SOLO_OS_ADDR: "@SOLO",
    OTHER_OS_ADDR: "@OTHEROS",
}

# File types
FILE_TYPE_SEGMENT = -1  # Segment
FILE_TYPE_EMPTY = 0  #    Empty file
FILE_TYPE_SCRATCH = 1  #  Scratch file
FILE_TYPE_ASCII = 2  #    Ascii file
FILE_TYPE_SEQCODE = 3  #  Sequential Pascal code file
FILE_TYPE_CONCODE = 4  #  Concurrent Pascal code file

FILE_TYPES = {
    FILE_TYPE_EMPTY: "EMPTY",
    FILE_TYPE_SCRATCH: "SCRATCH",
    FILE_TYPE_ASCII: "ASCII",
    FILE_TYPE_SEQCODE: "SEQCODE",
    FILE_TYPE_CONCODE: "CONCODE",
    FILE_TYPE_SEGMENT: "SEGMENT",
}


def get_file_type_id(file_type: t.Optional[str], default: int = FILE_TYPE_ASCII) -> int:
    """
    Get the file type id from a string
    """
    if not file_type:
        return default
    file_type = file_type.upper()
    for file_id, file_str in FILE_TYPES.items():
        if file_str == file_type:
            return file_id
    raise Exception("?KMON-F-Invalid file type specified with option")


def filename_hash(filename: str, catalog_length: int) -> int:
    """
    Calculate the hash for a given file name
    """
    key = 1
    for c in filename[:ID_LENGTH]:
        if c != " ":
            key = key * ord(c.upper()) % (catalog_length * CAT_PAGE_LENGTH) + 1
    return key


def solo_canonical_filename(fullname: t.Optional[str], wildcard: bool = False, segment: bool = False) -> str:
    """
    Generate the canonical SOLO name
    """

    def filter_fn(s: str) -> bool:
        if wildcard and s == "*":
            return True
        return str.isalnum(s)

    if segment and fullname and fullname.startswith("@"):
        return "@" + "".join(filter(filter_fn, fullname or "")).upper()[:ID_LENGTH]
    else:
        return "".join(filter(filter_fn, fullname or "")).upper()[:ID_LENGTH]


def solo_to_ascii(data: bytes) -> bytes:
    if not data:
        return data
    if not EM in data:
        data = data.rstrip(b"\0")
    else:
        data = data.split(EM)[0]
    return data


def ascii_to_solo(data: bytes) -> bytes:
    if not data:
        return data
    if not data.endswith(EM):
        data += EM
    return data


class SOLOFile(AbstractFile):
    entry: "SOLODirectoryEntry"
    file_mode: str
    closed: bool

    def __init__(self, entry: "SOLODirectoryEntry", file_mode: t.Optional[str] = None) -> None:
        self.entry = entry
        if file_mode:
            self.file_mode = file_mode
        else:
            self.file_mode = ASCII if self.entry.file_type_id == FILE_TYPE_ASCII else IMAGE
        self.closed = False

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.length
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.length
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        data = bytearray()
        for i in range(block_number, block_number + number_of_blocks):
            disk_block_number = self.entry.page_map[i]
            data.extend(self.entry.fs.read_block(disk_block_number))
        if data and self.file_mode == ASCII:
            return solo_to_ascii(bytes(data))
        else:
            return bytes(data)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.length
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        for i in range(block_number, block_number + number_of_blocks):
            disk_block_number = self.entry.page_map[i]
            t = buffer[i * BLOCK_SIZE : (i + 1) * BLOCK_SIZE]
            if len(t) < BLOCK_SIZE:
                t = t + bytes([0] * (BLOCK_SIZE - len(t)))  # Pad with zeros
            self.entry.fs.write_block(t, disk_block_number)

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.entry.length * BLOCK_SIZE

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class SOLOSegment(AbstractFile):
    """
    Segments are 3 pre-defined, contiguous, fixed-size files

             start    end      size
             block    block
    @KERNEL      0 => 23        24
    @SOLO       24 => 87        64
    @OTHEROS    88 => 152       64
    """

    entry: "SOLOSegmentDirectoryEntry"
    closed: bool

    def __init__(self, entry: "SOLOSegmentDirectoryEntry"):
        self.closed = False
        self.entry = entry

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the segment
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.length
        if self.closed or block_number < 0 or number_of_blocks < 0:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        if block_number + number_of_blocks > self.entry.size:
            number_of_blocks = self.entry.size - block_number
        return self.entry.fs.read_block(self.entry.segment_addr + block_number, number_of_blocks)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the segment
        """
        if self.closed or block_number < 0 or number_of_blocks < 0 or block_number + number_of_blocks > self.entry.size:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        self.entry.fs.write_block(buffer, self.entry.segment_addr + block_number, number_of_blocks)

    def get_size(self) -> int:
        """
        Get segment size in bytes
        """
        return self.entry.size * BLOCK_SIZE

    def get_block_size(self) -> int:
        """
        Get segment block size in bytes
        """
        return BLOCK_SIZE

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True


class SOLOAbstractSortableDirectoryEntry(AbstractDirectoryEntry):
    filename: str = ""

    @property
    def fullname(self) -> str:
        return self.filename

    @property
    def basename(self) -> str:
        return self.filename

    def __lt__(self, other: "SOLOAbstractSortableDirectoryEntry") -> bool:
        return self.filename < other.filename

    def __gt__(self, other: "SOLOAbstractSortableDirectoryEntry") -> bool:
        return self.filename > other.filename


class SOLODirectoryEntry(SOLOAbstractSortableDirectoryEntry):
    """
    SOLO Directory Entry

        +-------------------------------------+
     0  |               File                  |
    11  |               name                  |
        +-------------------------------------+
    12  |             File type               |
        +-------------------------------------+
    14  |       Page map block number         |
        +-------------------------------------+
    16  |           Protected flag            |
        +-------------------------------------+
    18  |              Spare                  |
        +-------------------------------------+
    28  |        Key (Filename hash)          |
        +-------------------------------------+
    30  |          Search length              |
        +-------------------------------------+

    """

    cat_page: "SOLOCatalogPage"
    filename: str = ""  # Filename (12 chars)
    file_type_id: int = 0  # File type (scratch, ascii, seqcode and concode)
    page_map_block_number: int = 0  # Page map block number
    protected: bool = False  # Protected against accidental overwriting or deletion
    spare: bytes = b""
    hash_key: int = 0  # Filename hash
    searchlength: int = 0  # Number of files with the same key
    page_map: t.List[int]  # Map file blocks to disk blocks

    def __init__(self, cat_page: "SOLOCatalogPage"):
        self.cat_page = cat_page

    @classmethod
    def new(
        cls,
        cat_page: "SOLOCatalogPage",
        filename: str,
        file_type_id: int,
        page_map_block_number: int,
        page_map: t.List[int],
        protected: bool,
        searchlength: int,
    ) -> "SOLODirectoryEntry":
        self = SOLODirectoryEntry(cat_page)
        self.filename = filename
        self.file_type_id = file_type_id
        self.page_map_block_number = page_map_block_number
        self.protected = protected
        self.hash_key = filename_hash(filename, cat_page.fs.catalog_length)
        self.searchlength = searchlength
        self.page_map = page_map
        return self

    @classmethod
    def read(cls, cat_page: "SOLOCatalogPage", buffer: bytes, position: int) -> "SOLODirectoryEntry":
        self = SOLODirectoryEntry(cat_page)
        (
            file_id,
            self.file_type_id,
            self.page_map_block_number,
            raw_protected,
            self.spare,
            self.hash_key,
            self.searchlength,
        ) = struct.unpack_from(ENTRY_FORMAT, buffer, position)
        self.filename = file_id.decode("ascii", errors="ignore").rstrip(" \x00")
        self.protected = bool(raw_protected)
        if self.is_empty:
            self.page_map = []
        else:
            self.page_map = self.cat_page.fs.read_page_map(self.page_map_block_number)
        return self

    def write_buffer(self, buffer: bytearray, position: int) -> None:
        file_id = self.filename.encode("ascii", errors="ignore").ljust(ID_LENGTH)
        raw_protected = 1 if self.protected else 0
        struct.pack_into(
            ENTRY_FORMAT,
            buffer,
            position,
            file_id,
            self.file_type_id,
            self.page_map_block_number,
            raw_protected,
            self.spare,
            self.hash_key,
            self.searchlength,
        )

    @property
    def is_empty(self) -> bool:
        return self.filename == ""

    @property
    def file_type(self) -> str:
        return FILE_TYPES.get(self.file_type_id, "")

    @property
    def length(self) -> int:
        """
        File length in blocks
        """
        return len(self.page_map)

    def get_length(self) -> int:
        """
        File length in blocks
        """
        return len(self.page_map)

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.length * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE

    @property
    def fs(self) -> "SOLOFilesystem":
        return self.cat_page.fs

    def delete(self) -> bool:
        # Update the free block bitmap
        bitmap = SOLOBitmap.read(self.fs)
        for block_number in self.page_map:
            bitmap.set_free(block_number)  # Mark block as free
        bitmap.set_free(self.page_map_block_number)  # Mark block as free
        bitmap.write()
        # Delete entry from the catalog
        old_key = self.hash_key
        self.hash_key = 0
        self.file_type_id = FILE_TYPE_EMPTY
        self.protected = False
        self.filename = ""
        self.page_map_block_number = 0
        self.cat_page.write()
        # Decrement the counter of files with the same key
        self.fs.update_searchlength(old_key, -1)
        return True

    def write(self) -> bool:
        """
        Write the directory entry
        """
        raise OSError(errno.EINVAL, "Invalid operation on SOLO filesystem")

    def open(self, file_mode: t.Optional[str] = None) -> SOLOFile:
        """
        Open a file
        """
        return SOLOFile(self, file_mode)

    def __str__(self) -> str:
        return f"{self.filename:<12}  \
{self.file_type:<8}  \
{'PROT' if self.protected else '    '}  \
Key: {self.hash_key:>6} {'('+str(self.searchlength)+')':<6}  \
Length: {self.length:>4}  \
Map: {self.page_map_block_number:>4}  \
Blocks: {[x for x in self.page_map]}"

    def __repr__(self) -> str:
        return str(self)


class SOLOSegmentDirectoryEntry(SOLOAbstractSortableDirectoryEntry):
    fs: "SOLOFilesystem"
    filename: str = ""
    segment_addr: int = 0
    file_type_id: int = FILE_TYPE_SEGMENT
    protected: bool = True
    is_empty: bool = False
    file_type: str = "SEGMENT"
    creation_date: t.Optional[date] = None

    def __init__(self, fs: "SOLOFilesystem", filename: str):
        self.fs = fs
        self.filename = filename.upper()
        self.segment_addr = -1
        for addr, segment_name in SEGMENTS.items():
            if segment_name == self.filename:
                self.segment_addr = addr
                break
        if self.segment_addr == -1:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)
        if self.segment_addr == KERNEL_ADDR:
            self.size = KERNEL_LENGTH
        else:
            self.size = SEGMENT_LENGTH

    @property
    def length(self) -> int:
        """
        Segment length in blocks
        """
        if self.segment_addr == KERNEL_ADDR:
            return KERNEL_LENGTH
        else:
            return SEGMENT_LENGTH

    def get_length(self) -> int:
        """
        File length in blocks
        """
        if self.segment_addr == KERNEL_ADDR:
            return KERNEL_LENGTH
        else:
            return SEGMENT_LENGTH

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.length * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE

    def delete(self) -> bool:
        return False

    def write(self) -> bool:
        """
        Write the directory entry
        """
        raise OSError(errno.EINVAL, "Invalid operation on segment")

    def open(self, file_mode: t.Optional[str] = None) -> SOLOSegment:
        """
        Open a segment
        """
        return SOLOSegment(self)

    def __str__(self) -> str:
        return f"{self.filename:<12}  {self.file_type:<8}  Length: {self.length:>4}"

    def __repr__(self) -> str:
        return str(self)


class SOLOBitmap:

    fs: "SOLOFilesystem"
    bitmaps: t.List[int]  # 2(blocks) x 30(groups) x 120bit integer

    def __init__(self, fs: "SOLOFilesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "SOLOFilesystem") -> "SOLOBitmap":
        """
        Read the bitmap blocks from the disk
        """
        self = SOLOBitmap(fs)
        self.bitmaps = []
        # Read the bitmaps from the disk
        for block_number in range(FREE_LIST_ADDR, FREE_LIST_ADDR + FREE_LIST_LENGTH):
            t = self.fs.read_block(block_number)
            # print(block_number, t)
            if not t:
                raise OSError(errno.EIO, f"Failed to read block {block_number}")
            for i in range(0, FREE_PAGE_LENGTH):
                position = i * (FREE_PAGE_GROUP_LENGTH + FREE_PAGE_GROUP_PAD)
                bitmap = int.from_bytes(t[position : position + FREE_PAGE_GROUP_LENGTH], byteorder="big", signed=False)
                self.bitmaps.append(bitmap)
            # TODO misc (0=> # of free blocks, 1=> first free block?)
            # misc = struct.unpack_from(FREE_PAGE_MISC_FORMAT, t, BLOCK_SIZE - FREE_PAGE_MISC_LENGTH)
            # print(">>>", block_number, misc, self.is_free(misc[1]))
        return self

    def write(self) -> None:
        """
        Write the bitmap blocks to the disk
        """
        free = self.free()
        first_free_block = self.find_first_free()
        bitmaps = list(self.bitmaps)
        for block_number in range(FREE_LIST_ADDR, FREE_LIST_ADDR + FREE_LIST_LENGTH):
            t = bytearray(BLOCK_SIZE)
            for i in range(0, FREE_PAGE_LENGTH):
                position = i * (FREE_PAGE_GROUP_LENGTH + FREE_PAGE_GROUP_PAD)
                bitmap = bitmaps.pop(0)
                t[position : position + FREE_PAGE_GROUP_LENGTH] = bitmap.to_bytes(
                    FREE_PAGE_GROUP_LENGTH, byteorder="big"
                )
            # Update first free block and number of free blocks
            struct.pack_into(FREE_PAGE_MISC_FORMAT, t, BLOCK_SIZE - FREE_PAGE_MISC_LENGTH, free, first_free_block)
            self.fs.write_block(bytes(t), block_number)

    @property
    def total_bits(self) -> int:
        """
        Return the bitmap length in bit
        """
        return DISK_SIZE

    def is_free(self, block_number: int) -> bool:
        """
        Check if a block is free
        """
        int_index = block_number // GROUP_LENGTH
        bit_position = block_number % GROUP_LENGTH
        bit_value = self.bitmaps[int_index]
        return (bit_value & (1 << bit_position)) != 0

    def set_free(self, block_number: int) -> None:
        """
        Mark a block as free
        """
        int_index = block_number // GROUP_LENGTH
        bit_position = block_number % GROUP_LENGTH
        self.bitmaps[int_index] |= 1 << bit_position

    def set_used(self, block_number: int) -> None:
        """
        Allocate a block
        """
        int_index = block_number // GROUP_LENGTH
        bit_position = block_number % GROUP_LENGTH
        self.bitmaps[int_index] &= ~(1 << bit_position)

    def allocate(self, size: int) -> t.List[int]:
        """
        Allocate blocks
        """
        blocks = []
        for block in range(0, self.total_bits):
            if self.is_free(block):
                self.set_used(block)  # Mark as used
                blocks.append(block)
            if len(blocks) == size:
                break
        if len(blocks) < size:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        return blocks

    def find_first_free(self) -> int:
        """
        Get the first free block (does not allocate the block)
        """
        for block in range(0, self.total_bits):
            if self.is_free(block):
                return block
        return self.total_bits

    def used(self) -> int:
        """
        Count the number of used blocks
        """
        return self.total_bits - self.free()

    def free(self) -> int:
        """
        Count the number of free blocks
        """
        free = 0
        for i, bitmap in enumerate(self.bitmaps):
            if i * GROUP_LENGTH < DISK_SIZE:  # Consider only the blocks < DISK_SIZE
                free += bitmap.bit_count()
        return free


class SOLOCatalogPage:
    """
    SOLO Catalog page

        +-------------------------------------+
     0  |          Directory Entry 1          |
        +-------------------------------------+
        |                 ...                 |
        +-------------------------------------+
    480 |          Directory Entry 16         |
        +-------------------------------------+
    """

    fs: "SOLOFilesystem"
    block_number: int  # Catalog block number
    entries: t.List["SOLODirectoryEntry"]  # Directory entries

    def __init__(self, fs: "SOLOFilesystem", block_number: int):
        self.fs = fs
        self.block_number = block_number

    @classmethod
    def read(cls, fs: "SOLOFilesystem", block_number: int) -> "SOLOCatalogPage":
        """
        Read a catalog page from the disk
        """
        self = SOLOCatalogPage(fs, block_number)
        buffer = self.fs.read_block(block_number)
        self.entries = [SOLODirectoryEntry.read(self, buffer, pos) for pos in range(0, BLOCK_SIZE, ENTRY_LENGTH)]
        return self

    def write(self) -> None:
        """
        Write the catalog page on the disk
        """
        buffer = bytearray(BLOCK_SIZE)
        for i, entry in enumerate(self.entries):
            pos = i * ENTRY_LENGTH
            entry.write_buffer(buffer, pos)
        self.fs.write_block(buffer, self.block_number)

    def create_entry(
        self,
        filename: str,
        file_type: t.Optional[str],
        page_map_block_number: int,
        page_map: t.List[int],
        search_key: t.Optional[int] = None,
        protected: bool = False,
    ) -> t.Optional["SOLODirectoryEntry"]:
        """
        Create a new entry in this catalog page.
        If search_key is not None, put the new entry in the first position after the search_key
        """
        file_type_id = get_file_type_id(file_type)
        if file_type_id == FILE_TYPE_EMPTY or file_type_id == FILE_TYPE_SEGMENT:
            raise Exception("?KMON-F-Invalid file type specified with option")
        if search_key is not None:
            pos = (search_key - 1) % CAT_PAGE_LENGTH
            for i, entry in enumerate(self.entries[pos:], start=pos):
                if entry.is_empty:
                    searchlength = self.entries[i].searchlength
                    self.entries[i] = SOLODirectoryEntry.new(
                        self, filename, file_type_id, page_map_block_number, page_map, protected, searchlength
                    )
                    return self.entries[i]
        else:
            for i, entry in enumerate(self.entries):
                if entry.is_empty:
                    searchlength = self.entries[i].searchlength
                    self.entries[i] = SOLODirectoryEntry.new(
                        self, filename, file_type_id, page_map_block_number, page_map, protected, searchlength
                    )
                    return self.entries[i]
        return None

    def __str__(self) -> str:
        return f"Catalog#{self.block_number}"

    def __repr__(self) -> str:
        return str(self)


class SOLOFilesystem(AbstractRXBlockFilesystem):
    """
    SOLO Filesystem

    P. Brinch Hansen,
    The Solo operating system: a Concurrent Pascal progran Software—Practice and Experience, 1976

    http://brinch-hansen.net/papers/1976b.pdf

    Block
          +-------------------------------------+
       0  |          Kernel (24 block)          |
          +-------------------------------------+
      24  |         Solo OS (64 blocks)         |
          +-------------------------------------+
      88  |        Other OS (64 blocks)         |
          +-------------------------------------+
     152  |    Free block bitmap (2 blocks)     |
          +-------------------------------------+
     154  |   Catalog pages index (1 blocks)    |
          +-------------------------------------+
          |           Catalog pages             |
          |         Directory entries           |
          |               Files                 |
          +-------------------------------------+
    """

    fs_name = "solo"
    fs_description = "PDP-11 SOLO"
    fs_platforms = ["pdp11"]

    catalog_length: int

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        **kwargs: t.Union[bool, str],
    ) -> "SOLOFilesystem":
        self = cls(file_or_dev)
        # Get catalog length
        buffer = self.read_block(CAT_ADDR)
        self.catalog_length = struct.unpack_from("<H", buffer, 0)[0]
        if self.catalog_length != 15:
            raise OSError(errno.EIO, "Invalid catalog length")
        return self

    def read_page_map(self, block_number: int) -> t.List[int]:
        """
        Read a page map
        """
        buffer = self.read_block(block_number)
        words = struct.unpack_from("<256H", buffer, 0)
        length = words[0]
        return list(words[1 : length + 1])

    def write_page_map(self, page_map: t.List[int], block_number: int) -> None:
        """
        Write a page map
        """
        if len(page_map) > MAX_FILE_SIZE:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        length = len(page_map)
        words = [length] + page_map
        if len(words) < 256:
            words += [0] * (256 - len(words))
        assert len(words) == 256
        buffer = struct.pack("<256H", *words)
        self.write_block(buffer, block_number)

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = True,
        expand: bool = True,
        wildcard: bool = True,
    ) -> t.Iterator[t.Union["SOLODirectoryEntry", "SOLOSegmentDirectoryEntry"]]:
        file_type_id: t.Optional[int] = None
        if pattern:
            if ";" in pattern:
                pattern, file_type = pattern.split(";", 1)
                file_type_id = get_file_type_id(file_type)
            pattern = solo_canonical_filename(pattern, segment=True, wildcard=True)
        if include_all or file_type_id == FILE_TYPE_SEGMENT:
            for segment in SEGMENTS.values():
                if filename_match(segment, pattern, wildcard):
                    if file_type_id is None or file_type_id == FILE_TYPE_SEGMENT:
                        yield SOLOSegmentDirectoryEntry(self, segment)

        for entry in self.entries_list:
            if filename_match(entry.basename, pattern, wildcard) and not entry.is_empty:
                if file_type_id is None or file_type_id == entry.file_type_id:
                    yield entry

    @property
    def entries_list(self) -> t.Iterator["SOLODirectoryEntry"]:
        """
        Read the catalog
        """
        for block_number in self.read_page_map(CAT_ADDR):
            catalog_page = SOLOCatalogPage.read(self, block_number)
            for entry in catalog_page.entries:
                yield entry

    def get_searchlength(self, hash_key: int) -> int:
        page_num = (hash_key - 1) // CAT_PAGE_LENGTH + 1
        pages = self.read_page_map(CAT_ADDR)
        block_number = pages[page_num - 1]
        catalog_page = SOLOCatalogPage.read(self, block_number)
        pos = (hash_key - 1) % CAT_PAGE_LENGTH
        entry = catalog_page.entries[pos]
        return entry.searchlength

    def update_searchlength(self, hash_key: int, delta: int) -> None:
        page_num = (hash_key - 1) // CAT_PAGE_LENGTH + 1
        pages = self.read_page_map(CAT_ADDR)
        block_number = pages[page_num - 1]
        catalog_page = SOLOCatalogPage.read(self, block_number)
        pos = (hash_key - 1) % CAT_PAGE_LENGTH
        entry = catalog_page.entries[pos]
        entry.searchlength += delta
        if entry.searchlength < 0:
            entry.searchlength = 0
        catalog_page.write()

    def get_first_file_entry_for_hash(self, hash_key: int) -> t.Optional[SOLODirectoryEntry]:
        page_num = (hash_key - 1) // CAT_PAGE_LENGTH + 1
        pages = self.read_page_map(CAT_ADDR)
        block_number = pages[page_num - 1]
        catalog_page = SOLOCatalogPage.read(self, block_number)
        for entry in catalog_page.entries:
            if entry.hash_key == hash_key:
                return entry
        return None

    def get_file_entry(self, fullname: str) -> t.Union[SOLODirectoryEntry, SOLOSegmentDirectoryEntry]:
        """
        Get the directory entry for a file
        """
        fullname = solo_canonical_filename(fullname, segment=True)
        if not fullname:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)
        if fullname.startswith("@"):
            return SOLOSegmentDirectoryEntry(self, fullname)
        # Lookup file by key hash
        hash_key = filename_hash(fullname, self.catalog_length)
        entry = self.get_first_file_entry_for_hash(hash_key)
        if entry is not None:
            # Check if the entry match by filename
            if entry.basename == fullname:
                return entry
            # Check the other entries in the same catalog page
            for entry in entry.cat_page.entries:
                if entry.basename == fullname:
                    return entry
        # Fallback
        try:
            return next(self.filter_entries_list(fullname, wildcard=False))
        except StopIteration:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
        protected: bool = False,
    ) -> None:
        """
        Write content to a file/segment
        """
        number_of_blocks = int(math.ceil(len(content) * 1.0 / BLOCK_SIZE))
        if number_of_blocks > MAX_FILE_SIZE:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        if get_file_type_id(file_type) == FILE_TYPE_ASCII:
            content = ascii_to_solo(content)

        entry = self.create_file(
            fullname=fullname, number_of_blocks=number_of_blocks, file_type=file_type, protected=protected
        )
        if isinstance(entry, SOLOSegmentDirectoryEntry):
            f: t.Union[SOLOSegment, SOLOFile] = SOLOSegment(entry)
        else:
            f = SOLOFile(entry)
        try:
            f.write_block(content, 0, entry.length)
        finally:
            f.close()

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
        protected: bool = False,
    ) -> t.Union[SOLODirectoryEntry, SOLOSegmentDirectoryEntry]:
        """
        Create a new file with a given length in number of blocks
        """
        if number_of_blocks > MAX_FILE_SIZE:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        # Delete the existing file
        fullname = solo_canonical_filename(fullname, segment=True)
        try:
            old_entry = self.get_file_entry(fullname)
            if isinstance(old_entry, SOLOSegmentDirectoryEntry):
                return old_entry
            old_entry.delete()
        except FileNotFoundError:
            pass
        # Allocate the space for the page map and the the file
        bitmap = SOLOBitmap.read(self)
        blocks = bitmap.allocate(number_of_blocks + 1)
        page_map_block_number = blocks[0]
        file_blocks = blocks[1:]
        self.write_page_map(file_blocks, page_map_block_number)
        # Lookup catalog entry by key hash
        fullname = solo_canonical_filename(fullname, segment=False)
        hash_key = filename_hash(fullname, self.catalog_length)
        # Create the catalog entry
        page_num = (hash_key - 1) // CAT_PAGE_LENGTH + 1  # 1 >= page_num >= CAT_PAGE_LENGTH
        cat_pages = self.read_page_map(CAT_ADDR)
        block_number = cat_pages[page_num - 1]
        catalog_page = SOLOCatalogPage.read(self, block_number)
        # print(f"{hash_key=} {page_num+1=} {block_number=} {catalog_page=}")
        new_entry = catalog_page.create_entry(
            fullname,
            file_type,
            page_map_block_number,
            file_blocks,
            search_key=hash_key,
            protected=protected,
        )
        if new_entry is None:
            for block_number in self.read_page_map(CAT_ADDR):
                catalog_page = SOLOCatalogPage.read(self, block_number)
                new_entry = catalog_page.create_entry(
                    fullname,
                    file_type,
                    page_map_block_number,
                    file_blocks,
                    protected=protected,
                )
                if new_entry is not None:
                    break
        if new_entry is None:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        catalog_page.write()
        # Write bitmap
        bitmap.write()
        # Increment the counter of files with the same key
        self.update_searchlength(new_entry.hash_key, delta=+1)
        return new_entry

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        files = 0
        blocks = 0
        if not options.get("brief"):
            sys.stdout.write("SOLO SYSTEM FILES\n\n")
        for x in sorted(self.filter_entries_list(pattern, include_all=bool(options.get("full")), wildcard=True)):
            if options.get("brief"):
                sys.stdout.write(f"{x.filename}\n")
            else:
                sys.stdout.write(
                    f"{x.filename:<12} {x.file_type:<12} {'PROTECTED' if x.protected else 'UNPROTECTED':<12} {x.length:>6} PAGES\n"
                )
            blocks += x.length
            files += 1
        if options.get("brief"):
            return
        sys.stdout.write(f"{files:>5} ENTRIES\n{blocks:>5} PAGES\n")

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        if options.get("bitmap"):
            # Print the bitmap
            bitmap = SOLOBitmap.read(self)
            for i in range(0, bitmap.total_bits):
                sys.stdout.write(f"{i:>4} {'[ ]' if bitmap.is_free(i) else '[X]'}  ")
                if i % 16 == 15:
                    sys.stdout.write("\n")
        elif not arg:
            # Print the catalog
            for segment in SEGMENTS.values():
                segment_entry = SOLOSegmentDirectoryEntry(self, segment)
                sys.stdout.write(f" -  {segment_entry}\n")
            t = 1
            for page_num, block_number in enumerate(self.read_page_map(CAT_ADDR), start=1):
                catalog_page = SOLOCatalogPage.read(self, block_number)
                for entry in catalog_page.entries:
                    sys.stdout.write(f"{t:>3} {page_num:>2}# {entry}\n")
                    t += 1
        else:
            self.dump(arg)

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()

    @classmethod
    def initialize(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "SOLOFilesystem":
        self = cls(file_or_dev)
        # Zero disk
        empty_block = b"\0" * BLOCK_SIZE
        for block_number in range(0, DISK_SIZE):
            self.write_block(empty_block, block_number)
        # Mark the blocks allocated in the bitmap
        bitmap = SOLOBitmap.read(self)
        for i in range(0, CAT_ADDR + 1):
            bitmap.set_used(i)
        for i in range(CAT_ADDR + 1, DISK_SIZE):
            bitmap.set_free(i)
        bitmap.write()
        # Create the catalog
        self.catalog_length = 15
        bitmap = SOLOBitmap.read(self)
        catalog_pages = bitmap.allocate(self.catalog_length)
        bitmap.write()
        self.write_page_map(catalog_pages, CAT_ADDR)
        for block_number in catalog_pages:
            catalog_page = SOLOCatalogPage.read(self, block_number)
            catalog_page.write()
        # Create NEXT
        self.create_file(fullname="NEXT", number_of_blocks=255, file_type="SCRATCH", protected=True)
        return self

    def get_types(self) -> t.List[str]:
        """
        Get the list of the supported file types
        """
        return list(FILE_TYPES.values())
